import numpy as np
import torch
from scipy.special import gamma
import astrophot as ap
from utils import make_basic_sersic, make_basic_gaussian

######################################################################
# Util functions
######################################################################


def test_make_psf():

    target = make_basic_gaussian(x=10, y=10)
    target += make_basic_gaussian(x=40, y=40, rand=54321)

    assert np.all(
        np.isfinite(target.data.detach().cpu().numpy())
    ), "Target image should be finite after creation"


def test_conversions_units():

    # flux to sb
    # flux to sb
    assert (
        ap.utils.conversions.units.flux_to_sb(1.0, 1.0, 0.0) == 0
    ), "flux incorrectly converted to sb"

    # sb to flux
    assert ap.utils.conversions.units.sb_to_flux(1.0, 1.0, 0.0) == (
        10 ** (-1 / 2.5)
    ), "sb incorrectly converted to flux"

    # flux to mag no error
    assert (
        ap.utils.conversions.units.flux_to_mag(1.0, 0.0) == 0
    ), "flux incorrectly converted to mag (no error)"

    # flux to mag with error
    assert ap.utils.conversions.units.flux_to_mag(1.0, 0.0, fluxe=1.0) == (
        0.0,
        2.5 / np.log(10),
    ), "flux incorrectly converted to mag (with error)"

    # mag to flux no error:
    assert ap.utils.conversions.units.mag_to_flux(1.0, 0.0, mage=None) == (
        10 ** (-1 / 2.5)
    ), "mag incorrectly converted to flux (no error)"

    # mag to flux with error:
    for i in range(1):
        assert np.isclose(
            ap.utils.conversions.units.mag_to_flux(1.0, 0.0, mage=1.0)[i],
            (10 ** (-1.0 / 2.5), np.log(10) * (1.0 / 2.5) * 10 ** (-1.0 / 2.5))[i],
        ), "mag incorrectly converted to flux (with error)"

    # magperarcsec2 to mag with area A defined
    assert np.isclose(
        ap.utils.conversions.units.magperarcsec2_to_mag(1.0, a=None, b=None, A=1.0),
        (1.0 - 2.5 * np.log10(1.0)),
    ), "mag/arcsec^2 incorrectly converted to mag (area A given, a and b not defined)"

    # magperarcsec2 to mag with semi major and minor axes defined (a, and b)
    assert np.isclose(
        ap.utils.conversions.units.magperarcsec2_to_mag(1.0, a=1.0, b=1.0, A=None),
        (1.0 - 2.5 * np.log10(np.pi)),
    ), "mag/arcsec^2 incorrectly converted to mag (semi major/minor axes defined)"

    # mag to magperarcsec2 with area A defined
    assert np.isclose(
        ap.utils.conversions.units.mag_to_magperarcsec2(1.0, a=None, b=None, A=1.0, R=None),
        (1.0 + 2.5 * np.log10(1.0)),
    ), "mag incorrectly converted to mag/arcsec^2 (area A given)"

    # mag to magperarcsec2 with radius R given (assumes circular)
    assert np.isclose(
        ap.utils.conversions.units.mag_to_magperarcsec2(1.0, a=None, b=None, A=None, R=1.0),
        (1.0 + 2.5 * np.log10(np.pi)),
    ), "mag incorrectly converted to mag/arcsec^2 (radius R given)"

    # mag to magperarcsec2 with semi major and minor axes defined (a, and b)
    assert np.isclose(
        ap.utils.conversions.units.mag_to_magperarcsec2(1.0, a=1.0, b=1.0, A=None, R=None),
        (1.0 + 2.5 * np.log10(np.pi)),
    ), "mag incorrectly converted to mag/arcsec^2 (area A given)"

    # position angle PA to radians
    assert np.isclose(
        ap.utils.conversions.units.PA_shift_convention(1.0, unit="rad"),
        ((1.0 - (np.pi / 2)) % np.pi),
    ), "PA incorrectly converted to radians"

    # position angle PA to degrees
    assert np.isclose(
        ap.utils.conversions.units.PA_shift_convention(1.0, unit="deg"), ((1.0 - (180 / 2)) % 180)
    ), "PA incorrectly converted to degrees"


def test_conversion_functions():

    sersic_n = ap.utils.conversions.functions.sersic_n_to_b(1.0)
    # sersic I0 to flux - numpy
    assert np.isclose(
        ap.utils.conversions.functions.sersic_I0_to_flux_np(1.0, 1.0, 1.0, 1.0),
        (2 * np.pi * gamma(2)),
    ), "Error converting sersic central intensity to flux (np)"
    # sersic flux to I0 - numpy
    assert np.isclose(
        ap.utils.conversions.functions.sersic_flux_to_I0_np(1.0, 1.0, 1.0, 1.0),
        (1.0 / (2 * np.pi * gamma(2))),
    ), "Error converting sersic flux to central intensity (np)"

    # sersic Ie to flux - numpy
    assert np.isclose(
        ap.utils.conversions.functions.sersic_Ie_to_flux_np(1.0, 1.0, 1.0, 1.0),
        (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2)),
    ), "Error converting sersic effective intensity to flux (np)"

    # sersic flux to Ie - numpy
    assert np.isclose(
        ap.utils.conversions.functions.sersic_flux_to_Ie_np(1.0, 1.0, 1.0, 1.0),
        (1 / (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2))),
    ), "Error converting sersic flux to effective intensity (np)"

    # inverse sersic - numpy
    assert np.isclose(
        ap.utils.conversions.functions.sersic_inv_np(1.0, 1.0, 1.0, 1.0),
        (1.0 - (1.0 / sersic_n) * np.log(1.0)),
    ), "Error computing inverse sersic function (np)"

    # sersic I0 to flux - torch
    tv = torch.tensor([[1.0]], dtype=torch.float64)
    assert torch.allclose(
        torch.round(
            ap.utils.conversions.functions.sersic_I0_to_flux_np(tv, tv, tv, tv),
            decimals=7,
        ),
        torch.round(torch.tensor([[2 * np.pi * gamma(2)]]), decimals=7),
    ), "Error converting sersic central intensity to flux (torch)"

    # sersic flux to I0 - torch
    assert torch.allclose(
        torch.round(
            ap.utils.conversions.functions.sersic_flux_to_I0_np(tv, tv, tv, tv),
            decimals=7,
        ),
        torch.round(torch.tensor([[1.0 / (2 * np.pi * gamma(2))]]), decimals=7),
    ), "Error converting sersic flux to central intensity (torch)"

    # sersic Ie to flux - torch
    assert torch.allclose(
        torch.round(
            ap.utils.conversions.functions.sersic_Ie_to_flux_np(tv, tv, tv, tv),
            decimals=7,
        ),
        torch.round(
            torch.tensor([[2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2)]]),
            decimals=7,
        ),
    ), "Error converting sersic effective intensity to flux (torch)"

    # sersic flux to Ie - torch
    assert torch.allclose(
        torch.round(
            ap.utils.conversions.functions.sersic_flux_to_Ie_np(tv, tv, tv, tv),
            decimals=7,
        ),
        torch.round(
            torch.tensor([[1 / (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2))]]),
            decimals=7,
        ),
    ), "Error converting sersic flux to effective intensity (torch)"

    # inverse sersic - torch
    assert torch.allclose(
        torch.round(ap.utils.conversions.functions.sersic_inv_np(tv, tv, tv, tv), decimals=7),
        torch.round(torch.tensor([[1.0 - (1.0 / sersic_n) * np.log(1.0)]]), decimals=7),
    ), "Error computing inverse sersic function (torch)"



from .segmentation_map import *
from .center import center_of_mass, recursive_center_of_mass
from .construct_psf import gaussian_psf, moffat_psf
from .variance import auto_variance
from .PA import polar_decomposition

__all__ = (
    "center_of_mass",
    "recursive_center_of_mass",
    "gaussian_psf",
    "moffat_psf",
    "centroids_from_segmentation_map",
    "PA_from_segmentation_map",
    "q_from_segmentation_map",
    "windows_from_segmentation_map",
    "scale_windows",
    "filter_windows",
    "transfer_windows",
    "auto_variance",
    "polar_decomposition",
)



import numpy as np
from .conversions.functions import sersic_n_to_b


def sersic_np(R, n, Re, Ie):
    """Sersic 1d profile function, works more generally with numpy
    operations. In the event that impossible values are passed to the
    function it returns large values to guide optimizers away from
    such values.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        n: sersic index restricted to n > 0.36
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    if np.any(np.array([n, Re, Ie]) <= 0):
        return np.ones(len(R)) * 1e6
    bn = sersic_n_to_b(n)
    return Ie * np.exp(-bn * ((R / Re) ** (1 / n) - 1))


def gaussian_np(R, sigma, I0):
    """Gaussian 1d profile function, works more generally with numpy
    operations.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        sigma: standard deviation of the gaussian in the same units as R
        I0: central surface density
    """
    return (I0 / np.sqrt(2 * np.pi * sigma**2)) * np.exp(-0.5 * ((R / sigma) ** 2))


def exponential_np(R, Ie, Re):
    """Exponential 1d profile function, works more generally with numpy
    operations.

    Parameters:
        R: Radii array at which to evaluate the sersic function
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    return Ie * np.exp(-sersic_n_to_b(1.0) * (R / Re - 1.0))


def moffat_np(R, n, Rd, I0):
    """Moffat 1d profile function, works with numpy operations.

    Parameters:
        R: Radii tensor at which to evaluate the moffat function
        n: concentration index
        Rd: scale length in the same units as R
        I0: central surface density

    """
    return I0 / (1 + (R / Rd) ** 2) ** n


def nuker_np(R, Rb, Ib, alpha, beta, gamma):
    """Nuker 1d profile function, works with numpy functions

    Parameters:
        R: Radii tensor at which to evaluate the nuker function
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )


def ferrer_np(R, rout, alpha, beta, I0):
    """
    Modified Ferrer profile.

    Parameters
    ----------
    R : array_like
        Radial distance from the center.
    rout : float
        Outer radius of the profile.
    alpha : float
        Power-law index.
    beta : float
        Exponent for the modified Ferrer function.
    I0 : float
        Central intensity.

    Returns
    -------
    array_like
        The modified Ferrer profile evaluated at R.
    """
    return (R < rout) * I0 * ((1 - (np.clip(R, 0, rout) / rout) ** (2 - beta)) ** alpha)


def king_np(R, Rc, Rt, alpha, I0):
    """
    Empirical King profile.

    Parameters
    ----------
    R : array_like
        The radial distance from the center.
    Rc : float
        The core radius of the profile.
    Rt : float
        The truncation radius of the profile.
    alpha : float
        The power-law index of the profile.
    I0 : float
        The central intensity of the profile.

    Returns
    -------
    array_like
        The intensity at each radial distance.
    """
    beta = 1 / (1 + (Rt / Rc) ** 2) ** (1 / alpha)
    gamma = 1 / (1 + (R / Rc) ** 2) ** (1 / alpha)
    return (R < Rt) * I0 * ((np.clip(gamma, 0, 1) - beta) / (1 - beta)) ** alpha



import torch
import numpy as np

from ...errors import OptimizeStopFail, OptimizeStopSuccess


def hessian(J, W):
    return J.T @ (W.unsqueeze(1) * J)


def gradient(J, W, R):
    return J.T @ (W * R).unsqueeze(1)


def damp_hessian(hess, L):
    I = torch.eye(len(hess), dtype=hess.dtype, device=hess.device)
    D = torch.ones_like(hess) - I
    return hess * (I + D / (1 + L)) + L * I * torch.diag(hess)


def solve(hess, grad, L):
    hessD = damp_hessian(hess, L)  # (N, N)
    while True:
        try:
            h = torch.linalg.solve(hessD, grad)
            break
        except torch._C._LinAlgError:
            hessD = hessD + L * torch.eye(len(hessD), dtype=hessD.dtype, device=hessD.device)
            L = L * 2
    return hessD, h


def lm_step(x, data, model, weight, jacobian, ndf, L=1.0, Lup=9.0, Ldn=11.0):
    M0 = model(x)  # (M,)
    J = jacobian(x)  # (M, N)
    R = data - M0  # (M,)
    chi20 = torch.sum(weight * R**2).item() / ndf
    grad = gradient(J, weight, R)  # (N, 1)
    hess = hessian(J, weight)  # (N, N)
    if torch.allclose(grad, torch.zeros_like(grad)):
        raise OptimizeStopSuccess("Gradient is zero, optimization converged.")

    best = {"x": torch.zeros_like(x), "chi2": chi20, "L": L}
    scary = {"x": None, "chi2": chi20, "L": L}

    nostep = True
    improving = None
    for _ in range(10):
        hessD, h = solve(hess, grad, L)  # (N, N), (N, 1)
        M1 = model(x + h.squeeze(1))  # (M,)
        chi21 = torch.sum(weight * (data - M1) ** 2).item() / ndf

        # Handle nan chi2
        if not np.isfinite(chi21):
            L *= Lup
            if improving is True:
                break
            improving = False
            continue

        if chi21 < scary["chi2"]:
            scary = {"x": x + h.squeeze(1), "chi2": chi21, "L": L}

        if torch.allclose(h, torch.zeros_like(h)) and L < 0.1:
            raise OptimizeStopSuccess("Step with zero length means optimization complete.")

        # actual chi2 improvement vs expected from linearization
        rho = (chi20 - chi21) * ndf / torch.abs(h.T @ hessD @ h - 2 * grad.T @ h).item()
        # Avoid highly non-linear regions
        if rho < 0.1 or rho > 2:
            L *= Lup
            if improving is True:
                break
            improving = False
            continue

        if chi21 < best["chi2"]:  # new best
            best = {"x": x + h.squeeze(1), "chi2": chi21, "L": L}
            nostep = False
            L /= Ldn
            if L < 1e-8 or improving is False:
                break
            improving = True
        elif improving is True:  # were improving, now not improving
            break
        else:  # not improving and bad chi2, damp more
            L *= Lup
            if L >= 1e9:
                break
            improving = False

        # If we are improving chi2 by more than 10% then we can stop
        if (best["chi2"] - chi20) / chi20 < -0.1:
            break

    if nostep:
        if scary["x"] is not None:
            return scary
        raise OptimizeStopFail("Could not find step to improve chi^2")

    return best



import torch
import numpy as np

import astrophot as ap
from utils import make_basic_sersic
import pytest

######################################################################
# Fit Objects
######################################################################


@pytest.mark.parametrize("center", [[20, 20], [25.1, 17.324567]])
@pytest.mark.parametrize("PA", [0, 60 * np.pi / 180])
@pytest.mark.parametrize("q", [0.2, 0.8])
@pytest.mark.parametrize("n", [1, 4])
@pytest.mark.parametrize("Re", [10, 25.1])
def test_chunk_jacobian(center, PA, q, n, Re):
    target = make_basic_sersic()
    model = ap.Model(
        name="test sersic",
        model_type="sersic galaxy model",
        center=center,
        PA=PA,
        q=q,
        n=n,
        Re=Re,
        Ie=10.0,
        target=target,
        integrate_mode="none",
    )

    Jtrue = model.jacobian()

    model.jacobian_maxparams = 3

    Jchunked = model.jacobian()
    assert torch.allclose(
        Jtrue.data, Jchunked.data
    ), "Param chunked Jacobian should match full Jacobian"

    model.jacobian_maxparams = 10
    model.jacobian_maxpixels = 20**2

    Jchunked = model.jacobian()

    assert torch.allclose(
        Jtrue.data, Jchunked.data
    ), "Pixel chunked Jacobian should match full Jacobian"


@pytest.fixture
def sersic_model():
    target = make_basic_sersic()
    model = ap.Model(
        name="test sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=np.pi,
        q=0.7,
        n=2,
        Re=15,
        Ie=10.0,
        target=target,
    )
    model.initialize()
    return model


@pytest.mark.parametrize(
    "fitter", [ap.fit.LM, ap.fit.Grad, ap.fit.ScipyFit, ap.fit.MHMCMC, ap.fit.MiniFit]
)
def test_fitters(fitter, sersic_model):
    model = sersic_model
    model.initialize()
    ll_init = model.gaussian_log_likelihood()
    pll_init = model.poisson_log_likelihood()
    result = fitter(model, max_iter=100).fit()
    ll_final = model.gaussian_log_likelihood()
    pll_final = model.poisson_log_likelihood()
    assert ll_final > ll_init, f"{fitter.__name__} should improve the log likelihood"
    assert pll_final > pll_init, f"{fitter.__name__} should improve the poisson log likelihood"


def test_fitters_iter():
    target = make_basic_sersic()
    model1 = ap.Model(
        name="test1",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=np.pi,
        q=0.7,
        n=2,
        Re=15,
        Ie=10.0,
        target=target,
    )
    model2 = ap.Model(
        name="test2",
        model_type="sersic galaxy model",
        center=[20.5, 21],
        PA=1.5 * np.pi,
        q=0.9,
        n=1,
        Re=10,
        Ie=8.0,
        target=target,
    )
    model = ap.Model(
        name="test group",
        model_type="group model",
        models=[model1, model2],
        target=target,
    )
    model.initialize()
    ll_init = model.gaussian_log_likelihood()
    pll_init = model.poisson_log_likelihood()
    result = ap.fit.Iter(model, max_iter=10).fit()
    ll_final = model.gaussian_log_likelihood()
    pll_final = model.poisson_log_likelihood()
    assert ll_final > ll_init, f"Iter should improve the log likelihood"
    assert pll_final > pll_init, f"Iter should improve the poisson log likelihood"

    # test hessian
    Hgauss = model.hessian(likelihood="gaussian")
    assert torch.all(torch.isfinite(Hgauss)), "Hessian should be finite for Gaussian likelihood"
    Hpoisson = model.hessian(likelihood="poisson")
    assert torch.all(torch.isfinite(Hpoisson)), "Hessian should be finite for Poisson likelihood"


def test_hessian(sersic_model):
    model = sersic_model
    model.initialize()
    Hgauss = model.hessian(likelihood="gaussian")
    assert torch.all(torch.isfinite(Hgauss)), "Hessian should be finite for Gaussian likelihood"
    Hpoisson = model.hessian(likelihood="poisson")
    assert torch.all(torch.isfinite(Hpoisson)), "Hessian should be finite for Poisson likelihood"
    assert Hgauss is not None, "Hessian should be computed for Gaussian likelihood"
    assert Hpoisson is not None, "Hessian should be computed for Poisson likelihood"
    with pytest.raises(ValueError):
        model.hessian(likelihood="unknown")


def test_gradient(sersic_model):
    model = sersic_model
    target = model.target
    target.weight = 1 / (10 + target.variance.T)
    model.initialize()
    x = model.build_params_array()
    grad = model.gradient()
    assert torch.all(torch.isfinite(grad)), "Gradient should be finite"
    assert grad.shape == x.shape, "Gradient shape should match parameters shape"
    x.requires_grad = True
    ll = model.gaussian_log_likelihood(x)
    ll.backward()
    autograd = x.grad
    assert torch.allclose(grad, autograd, rtol=1e-4), "Gradient should match autograd gradient"

    funcgrad = torch.func.grad(model.gaussian_log_likelihood)(x)
    assert torch.allclose(grad, funcgrad, rtol=1e-4), "Gradient should match functional gradient"


# class TestHMC(unittest.TestCase):
#     def test_hmc_sample(self):
#         np.random.seed(12345)
#         N = 50
#         pixelscale = 0.8
#         true_params = {
#             "n": 2,
#             "Re": 10,
#             "Ie": 1,
#             "center": [-3.3, 5.3],
#             "q": 0.7,
#             "PA": np.pi / 4,
#         }
#         target = ap.image.Target_Image(
#             data=np.zeros((N, N)),
#             pixelscale=pixelscale,
#         )

#         MODEL = ap.models.Sersic_Galaxy(
#             name="sersic model",
#             target=target,
#             parameters=true_params,
#         )
#         img = MODEL().data.detach().cpu().numpy()
#         target.data = torch.Tensor(
#             img
#             + np.random.normal(scale=0.1, size=img.shape)
#             + np.random.normal(scale=np.sqrt(img) / 10)
#         )
#         target.variance = torch.Tensor(0.1**2 + img / 100)

#         HMC = ap.fit.HMC(MODEL, epsilon=1e-5, max_iter=5, warmup=2)
#         HMC.fit()


# class TestNUTS(unittest.TestCase):
#     def test_nuts_sample(self):
#         np.random.seed(12345)
#         N = 50
#         pixelscale = 0.8
#         true_params = {
#             "n": 2,
#             "Re": 10,
#             "Ie": 1,
#             "center": [-3.3, 5.3],
#             "q": 0.7,
#             "PA": np.pi / 4,
#         }
#         target = ap.image.Target_Image(
#             data=np.zeros((N, N)),
#             pixelscale=pixelscale,
#         )

#         MODEL = ap.models.Sersic_Galaxy(
#             name="sersic model",
#             target=target,
#             parameters=true_params,
#         )
#         img = MODEL().data.detach().cpu().numpy()
#         target.data = torch.Tensor(
#             img
#             + np.random.normal(scale=0.1, size=img.shape)
#             + np.random.normal(scale=np.sqrt(img) / 10)
#         )
#         target.variance = torch.Tensor(0.1**2 + img / 100)

#         NUTS = ap.fit.NUTS(MODEL, max_iter=5, warmup=2)
#         NUTS.fit()



from caskade import forward

from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from ..utils.conversions.functions import moffat_I0_to_flux
from .mixins import (
    MoffatMixin,
    InclinedMixin,
    RadialMixin,
    WedgeMixin,
    RayMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    iMoffatMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = (
    "MoffatGalaxy",
    "MoffatPSF",
    "Moffat2DPSF",
    "MoffatSuperEllipse",
    "MoffatFourierEllipse",
    "MoffatWarp",
    "MoffatRay",
    "MoffatWedge",
)


@combine_docstrings
class MoffatGalaxy(MoffatMixin, RadialMixin, GalaxyModel):
    usable = True

    @forward
    def total_flux(self, window=None, n=None, Rd=None, I0=None, q=None):
        return moffat_I0_to_flux(I0, n, Rd, q)


@combine_docstrings
class MoffatPSF(MoffatMixin, RadialMixin, PSFModel):
    _parameter_specs = {"I0": {"units": "flux/arcsec^2", "value": 1.0}}

    usable = True

    @forward
    def total_flux(self, window=None, n=None, Rd=None, I0=None):
        return moffat_I0_to_flux(I0, n, Rd, 1.0)


@combine_docstrings
class Moffat2DPSF(MoffatMixin, InclinedMixin, RadialMixin, PSFModel):

    _model_type = "2d"
    _parameter_specs = {"I0": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True


@combine_docstrings
class MoffatSuperEllipse(MoffatMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class MoffatFourierEllipse(MoffatMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class MoffatWarp(MoffatMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class MoffatRay(iMoffatMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class MoffatWedge(iMoffatMixin, WedgeMixin, GalaxyModel):
    usable = True



def moffat(R, n, Rd, I0):
    """Moffat 1d profile function

    Parameters:
        R: Radii tensor at which to evaluate the moffat function
        n: concentration index
        Rd: scale length in the same units as R
        I0: central surface density

    """
    return I0 / (1 + (R / Rd) ** 2) ** n



from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from .mixins import (
    SplineMixin,
    RadialMixin,
    iSplineMixin,
    RayMixin,
    WedgeMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = [
    "SplineGalaxy",
    "SplinePSF",
    "SplineWarp",
    "SplineSuperEllipse",
    "SplineFourierEllipse",
    "SplineRay",
    "SplineWedge",
]


@combine_docstrings
class SplineGalaxy(SplineMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SplinePSF(SplineMixin, RadialMixin, PSFModel):
    usable = True


@combine_docstrings
class SplineSuperEllipse(SplineMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SplineFourierEllipse(SplineMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SplineWarp(SplineMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SplineRay(iSplineMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class SplineWedge(iSplineMixin, WedgeMixin, GalaxyModel):
    usable = True



import torch

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import sersic_np
from .. import func


def _x0_func(model, R, F):
    return 2.0, R[4], 10 ** F[4]


class SersicMixin:
    """Sersic radial light profile (Sersic 1963).

    This is a classic profile used widely in galaxy modelling. It can be a good
    starting point for many extended objects. The functional form of the Sersic
    profile is defined as:

    $$I(R) = I_e * \\exp(- b_n((R/R_e)^(1/n) - 1))$$

    It is a generalization of a gaussian, exponential, and de-Vaucouleurs
    profile. The Sersic index `n` controls the shape of the profile, with `n=1`
    being an exponential profile, `n=4` being a de-Vaucouleurs profile, and
    `n=0.5` being a Gaussian profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Re: half light radius [arcsec]
        Ie: intensity at the half light radius [flux/arcsec^2]
    """

    _model_type = "sersic"
    _parameter_specs = {
        "n": {"units": "none", "valid": (0.36, 8), "shape": ()},
        "Re": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self, self.target[self.window], sersic_np, ("n", "Re", "Ie"), _x0_func
        )

    @forward
    def radial_model(self, R, n, Re, Ie):
        return func.sersic(R, n, Re, Ie)


class iSersicMixin:
    """Sersic radial light profile (Sersic 1963).

    This is a classic profile used widely in galaxy modelling. It can be a good
    starting point for many extended objects. The functional form of the Sersic
    profile is defined as:

    $$I(R) = I_e * \\exp(- b_n((R/R_e)^(1/n) - 1))$$

    It is a generalization of a gaussian, exponential, and de-Vaucouleurs
    profile. The Sersic index `n` controls the shape of the profile, with `n=1`
    being an exponential profile, `n=4` being a de-Vaucouleurs profile, and
    `n=0.5` being a Gaussian profile.

    `n`, `Re`, and `Ie` are batched by their first dimension, allowing for
    multiple Sersic profiles to be defined at once.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Re: half light radius [arcsec]
        Ie: intensity at the half light radius [flux/arcsec^2]
    """

    _model_type = "sersic"
    _parameter_specs = {
        "n": {"units": "none", "valid": (0.36, 8)},
        "Re": {"units": "arcsec", "valid": (0, None)},
        "Ie": {"units": "flux/arcsec^2", "valid": (0, None)},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=sersic_np,
            params=("n", "Re", "Ie"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, n, Re, Ie):
        return func.sersic(R, n[i], Re[i], Ie[i])



from functools import lru_cache
from scipy.special import binom
import numpy as np


@lru_cache(maxsize=1024)
def coefficients(n, m):
    C = []
    for k in range(int((n - abs(m)) / 2) + 1):
        C.append(
            (
                k,
                (-1) ** k * binom(n - k, k) * binom(n - 2 * k, (n - abs(m)) / 2 - k),
            )
        )
    return C


def zernike_n_m_list(n):
    nm = []
    for n_i in range(n + 1):
        for m_i in range(-n_i, n_i + 1, 2):
            nm.append((n_i, m_i))
    return nm


def zernike_n_m_modes(rho, phi, n, m):
    Z = np.zeros_like(rho)
    for k, c in coefficients(n, m):
        R = rho ** (n - 2 * k)
        T = 1.0
        if m < 0:
            T = np.sin(abs(m) * phi)
        elif m > 0:
            T = np.cos(m * phi)

        Z = Z + c * R * T
    return Z * (rho <= 1).astype(np.float64)



import astrophot as ap
import numpy as np
import torch
import pytest

######################################################################
# Image List Object
######################################################################


def test_image_creation():
    arr1 = torch.zeros((10, 15))
    base_image1 = ap.Image(data=arr1, pixelscale=1.0, zeropoint=1.0, name="image1")
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(data=arr2, pixelscale=0.5, zeropoint=2.0, name="image2")

    test_image = ap.ImageList((base_image1, base_image2))

    slicer = ap.WindowList(
        (ap.Window((3, 12, 5, 8), base_image1), ap.Window((4, 8, 3, 13), base_image2))
    )
    sliced_image = test_image[slicer]
    print(sliced_image[0].shape, sliced_image[1].shape)
    assert sliced_image[0].shape == (9, 3), "image slice incorrect shape"
    assert sliced_image[1].shape == (4, 10), "image slice incorrect shape"
    assert np.all(sliced_image[0].crpix == np.array([-3, -5])), "image should track origin"
    assert np.all(sliced_image[1].crpix == np.array([-4, -3])), "image should track origin"


def test_copy():
    arr1 = torch.zeros((10, 15)) + 2
    base_image1 = ap.Image(data=arr1, pixelscale=1.0, zeropoint=1.0, name="image1")
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(data=arr2, pixelscale=0.5, zeropoint=2.0, name="image2")

    test_image = ap.ImageList((base_image1, base_image2))

    copy_image = test_image.copy()
    copy_image.images[0] += 5
    copy_image.images[1] += 5

    for ti, ci in zip(test_image, copy_image):
        assert ti.pixelscale == ci.pixelscale, "copied image should have same pixelscale"
        assert ti.zeropoint == ci.zeropoint, "copied image should have same zeropoint"
        assert torch.all(ti.data != ci.data), "copied image should not modify original data"

    blank_copy_image = test_image.blank_copy()
    for ti, ci in zip(test_image, blank_copy_image):
        assert ti.pixelscale == ci.pixelscale, "copied image should have same pixelscale"
        assert ti.zeropoint == ci.zeropoint, "copied image should have same zeropoint"


def test_image_arithmetic():
    arr1 = torch.zeros((10, 15))
    base_image1 = ap.Image(data=arr1, pixelscale=1.0, zeropoint=1.0, name="image1")
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(data=arr2, pixelscale=0.5, zeropoint=2.0, name="image2")
    test_image = ap.ImageList((base_image1, base_image2))

    base_image3 = base_image1.copy()
    base_image3 += 1
    base_image4 = base_image2.copy()
    base_image4 -= 2
    second_image = ap.ImageList((base_image3, base_image4))

    # Test iadd
    test_image += second_image

    assert torch.allclose(
        test_image[0].data, torch.ones_like(base_image1.data)
    ), "image addition should update its region"
    assert torch.allclose(
        base_image1.data, torch.ones_like(base_image1.data)
    ), "image addition should update its region"
    assert torch.allclose(
        test_image[1].data, torch.zeros_like(base_image2.data)
    ), "image addition should update its region"
    assert torch.allclose(
        base_image2.data, torch.zeros_like(base_image2.data)
    ), "image addition should update its region"

    # Test isub
    test_image -= second_image

    assert torch.allclose(
        test_image[0].data, torch.zeros_like(base_image1.data)
    ), "image addition should update its region"
    assert torch.allclose(
        base_image1.data, torch.zeros_like(base_image1.data)
    ), "image addition should update its region"
    assert torch.allclose(
        test_image[1].data, torch.ones_like(base_image2.data)
    ), "image addition should update its region"
    assert torch.allclose(
        base_image2.data, torch.ones_like(base_image2.data)
    ), "image addition should update its region"

    new_image = test_image + second_image
    new_image = test_image - second_image
    new_image = new_image.to(dtype=torch.float32, device="cpu")
    assert isinstance(new_image, ap.ImageList), "new image should be an ImageList"

    new_image += base_image1
    new_image -= base_image2


def test_model_image_list_error():
    arr1 = torch.zeros((10, 15))
    base_image1 = ap.ModelImage(data=arr1, pixelscale=1.0, zeropoint=1.0)
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(data=arr2, pixelscale=0.5, zeropoint=2.0)

    with pytest.raises(ap.errors.InvalidImage):
        ap.ModelImageList((base_image1, base_image2))


def test_target_image_list_creation():
    arr1 = torch.zeros((10, 15))
    base_image1 = ap.TargetImage(
        data=arr1,
        pixelscale=1.0,
        zeropoint=1.0,
        variance=torch.ones_like(arr1),
        mask=torch.zeros_like(arr1),
        name="image1",
    )
    arr2 = torch.ones((15, 10))
    base_image2 = ap.TargetImage(
        data=arr2,
        pixelscale=0.5,
        zeropoint=2.0,
        variance=torch.ones_like(arr2),
        mask=torch.zeros_like(arr2),
        name="image2",
    )

    test_image = ap.TargetImageList((base_image1, base_image2))

    save_image = test_image.copy()
    second_image = test_image.copy()

    second_image[0].data += 1
    second_image[1].data += 1

    test_image += second_image
    test_image -= second_image

    assert torch.all(
        test_image[0].data == save_image[0].data
    ), "adding then subtracting should give the same image"
    assert torch.all(
        test_image[1].data == save_image[1].data
    ), "adding then subtracting should give the same image"


def test_targetlist_errors():
    arr1 = torch.zeros((10, 15))
    base_image1 = ap.TargetImage(
        data=arr1,
        pixelscale=1.0,
        zeropoint=1.0,
        variance=torch.ones_like(arr1),
        mask=torch.zeros_like(arr1),
    )
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(
        data=arr2,
        pixelscale=0.5,
        zeropoint=2.0,
    )
    with pytest.raises(ap.errors.InvalidImage):
        ap.TargetImageList((base_image1, base_image2))


def test_jacobian_image_list_error():
    arr1 = torch.zeros((10, 15, 3))
    base_image1 = ap.JacobianImage(
        parameters=["a", "1", "zz"], data=arr1, pixelscale=1.0, zeropoint=1.0
    )
    arr2 = torch.ones((15, 10))
    base_image2 = ap.Image(data=arr2, pixelscale=0.5, zeropoint=2.0)

    with pytest.raises(ap.errors.InvalidImage):
        ap.JacobianImageList((base_image1, base_image2))



import torch
import numpy as np

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import _sample_image
from ...utils.interpolate import default_prof
from .. import func


class SplineMixin:
    """Spline radial model for brightness.

    The `radial_model` function for this model is defined as a spline
    interpolation from the parameter `I_R`. The `I_R` parameter is a tensor
    that contains the radial profile of the brightness in units of
    flux/arcsec^2. The radius of each node is determined from `I_R.prof`.

    Parameters:
        I_R: Tensor of radial brightness values in units of flux/arcsec^2.
    """

    _model_type = "spline"
    _parameter_specs = {"I_R": {"units": "flux/arcsec^2", "valid": (0, None)}}

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I_R.initialized:
            return

        target_area = self.target[self.window]
        # Create the I_R profile radii if needed
        if self.I_R.prof is None:
            prof = default_prof(self.window.shape, target_area.pixelscale, 2, 0.2)
            self.I_R.prof = prof
        else:
            prof = self.I_R.prof

        R, I, S = _sample_image(
            target_area,
            self.transform_coordinates,
            self.radius_metric,
            rad_bins=[0] + list((prof[:-1] + prof[1:]) / 2) + [prof[-1] * 100],
        )
        self.I_R.dynamic_value = 10**I

    @forward
    def radial_model(self, R, I_R):
        return func.spline(R, self.I_R.prof, I_R)


class iSplineMixin:
    """Batched spline radial model for brightness.

    The `radial_model` function for this model is defined as a spline
    interpolation from the parameter `I_R`. The `I_R` parameter is a tensor that
    contains the radial profile of the brightness in units of flux/arcsec^2. The
    radius of each node is determined from `I_R.prof`.

    Both `I_R` and `I_R.prof` are batched by their first dimension, allowing for
    multiple spline profiles to be defined at once. Each individual spline model
    is then `I_R[i]` and `I_R.prof[i]` where `i` indexes the profiles.

    Parameters:
        I_R: Tensor of radial brightness values in units of flux/arcsec^2.
    """

    _model_type = "spline"
    _parameter_specs = {"I_R": {"units": "flux/arcsec^2", "valid": (0, None)}}

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I_R.initialized:
            return

        target_area = self.target[self.window]
        # Create the I_R profile radii if needed
        if self.I_R.prof is None:
            prof = default_prof(self.window.shape, target_area.pixelscale, 2, 0.2)
            prof = np.stack([prof] * self.segments)
            self.I_R.prof = prof
        else:
            prof = self.I_R.prof

        value = np.zeros(prof.shape)
        cycle = np.pi if self.symmetric else 2 * np.pi
        w = cycle / self.segments
        v = w * np.arange(self.segments)
        for s in range(self.segments):
            angle_range = (v[s] - w / 2, v[s] + w / 2)
            R, I, S = _sample_image(
                target_area,
                self.transform_coordinates,
                self.radius_metric,
                angle=self.angular_metric,
                rad_bins=[0] + list((prof[s][:-1] + prof[s][1:]) / 2) + [prof[s][-1] * 100],
                angle_range=angle_range,
                cycle=cycle,
            )
            value[s] = I

        self.I_R.dynamic_value = 10**value

    @forward
    def iradial_model(self, i, R, I_R):
        return func.spline(R, self.I_R.prof[i], I_R[i])



C1 = 4 / 405
C2 = 46 / 25515
C3 = 131 / 1148175
C4 = -2194697 / 30690717750


def sersic_n_to_b(n):
    """Compute the `b(n)` for a sersic model. This factor ensures that
    the :math:`R_e` and :math:`I_e` parameters do in fact correspond
    to the half light values and not some other scale
    radius/intensity.

    """
    x = 1 / n
    return 2 * n - 1 / 3 + x * (C1 + x * (C2 + x * (C3 + C4 * x)))


def sersic(R, n, Re, Ie):
    """Seric 1d profile function, specifically designed for pytorch
    operations

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        n: sersic index restricted to n > 0.36
        Re: Effective radius in the same units as R
        Ie: Effective surface density
    """
    bn = sersic_n_to_b(n)
    return Ie * (-bn * ((R / Re) ** (1 / n) - 1)).exp()



import astrophot as ap
import numpy as np


def test_window_creation():

    image = ap.Image(
        data=np.zeros((100, 110)),
        pixelscale=0.3,
        zeropoint=1.0,
        name="test_image",
    )
    window = ap.Window((2, 107, 3, 97), image)

    assert np.all(window.crpix == image.crpix), "Window should inherit crpix from image"
    assert window.identity == image.identity, "Window should inherit identity from image"
    assert window.shape == (105, 94), "Window should have correct shape"
    assert window.extent == (2, 107, 3, 97), "Window should have correct extent"
    assert str(window) == "Window(2, 107, 3, 97)", "String representation should match"


def test_window_chunk():

    image = ap.Image(
        data=np.zeros((100, 110)),
        pixelscale=0.3,
        zeropoint=1.0,
        name="test_image",
    )
    window1 = ap.Window((2, 107, 3, 97), image)

    subwindows = window1.chunk(10**2)
    reconstitute = subwindows[0]
    for subwindow in subwindows:
        reconstitute |= subwindow
    assert (
        reconstitute.i_low == window1.i_low
    ), "chunked windows should reconstitute to original window"
    assert (
        reconstitute.i_high == window1.i_high
    ), "chunked windows should reconstitute to original window"
    assert (
        reconstitute.j_low == window1.j_low
    ), "chunked windows should reconstitute to original window"
    assert (
        reconstitute.j_high == window1.j_high
    ), "chunked windows should reconstitute to original window"


def test_window_arithmetic():

    image = ap.Image(
        data=np.zeros((100, 110)),
        pixelscale=0.3,
        zeropoint=1.0,
        name="test_image",
    )
    windowbig = ap.Window((2, 107, 3, 97), image)
    windowsmall = ap.Window((20, 45, 30, 90), image)

    # Logical or, size
    ######################################################################
    big_or_small = windowbig | windowsmall
    assert big_or_small.i_low == 2, "logical or of images should take largest bounding box"
    assert big_or_small.i_high == 107, "logical or of images should take largest bounding box"
    assert big_or_small.j_low == 3, "logical or of images should take largest bounding box"
    assert big_or_small.j_high == 97, "logical or of images should take largest bounding box"

    # Logical and, size
    ######################################################################
    big_and_small = windowbig & windowsmall
    assert big_and_small.i_low == 20, "logical and of images should take overlap region"
    assert big_and_small.i_high == 45, "logical and of images should take overlap region"
    assert big_and_small.j_low == 30, "logical and of images should take overlap region"
    assert big_and_small.j_high == 90, "logical and of images should take overlap region"



from .base import AstroPhotError

__all__ = ("OptimizeStopFail", "OptimizeStopSuccess")


class OptimizeStopFail(AstroPhotError):
    """
    Raised at any point to stop optimization process due to failure.
    """

    pass


class OptimizeStopSuccess(AstroPhotError):
    """
    Raised at any point to stop optimization process due to success condition.
    """

    pass



import torch
import numpy as np

from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings
from . import func
from ..param import forward

__all__ = ["MultiGaussianExpansion"]


class MultiGaussianExpansion(ComponentModel):
    """Model that represents a galaxy as a sum of multiple Gaussian
    profiles. The model is defined as:

    I(R) = sum_i flux_i * exp(-0.5*(R_i / sigma_i)^2) / (2 * pi * q_i * sigma_i^2)

    where $R_i$ is a radius computed using $q_i$ and $PA_i$ for that component. All components share the same center.

    Parameters:
        q: axis ratio to scale minor axis from the ratio of the minor/major axis b/a, this parameter is unitless, it is restricted to the range (0,1)
        PA: position angle of the semi-major axis relative to the image positive x-axis in radians, it is a cyclic parameter in the range [0,pi)
        sigma: standard deviation of each Gaussian
        flux: amplitude of each Gaussian
    """

    _model_type = "mge"
    _parameter_specs = {
        "q": {"units": "b/a", "valid": (0, 1)},
        "PA": {"units": "radians", "valid": (0, np.pi), "cyclic": True},
        "sigma": {"units": "arcsec", "valid": (0, None)},
        "flux": {"units": "flux"},
    }
    usable = True

    def __init__(self, *args, n_components=None, **kwargs):
        super().__init__(*args, **kwargs)
        if n_components is None:
            for key in ("q", "sigma", "flux"):
                if self[key].value is not None:
                    self.n_components = self[key].value.shape[0]
                    break
            else:
                self.n_components = 1
        else:
            self.n_components = int(n_components)

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        target_area = self.target[self.window]
        dat = target_area.data.detach().cpu().numpy().copy()
        if target_area.has_mask:
            mask = target_area.mask.detach().cpu().numpy()
            dat[mask] = np.median(dat[~mask])
        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.nanmedian(edge)
        dat -= edge_average

        if not self.sigma.initialized:
            self.sigma.dynamic_value = np.logspace(
                np.log10(target_area.pixelscale.item() * 3),
                max(target_area.shape) * target_area.pixelscale.item() * 0.7,
                self.n_components,
            )
        if not self.flux.initialized:
            self.flux.dynamic_value = (np.sum(dat) / self.n_components) * np.ones(self.n_components)

        if self.PA.initialized or self.q.initialized:
            return

        x, y = target_area.coordinate_center_meshgrid()
        x = (x - self.center.value[0]).detach().cpu().numpy()
        y = (y - self.center.value[1]).detach().cpu().numpy()
        mu20 = np.median(dat * np.abs(x))
        mu02 = np.median(dat * np.abs(y))
        mu11 = np.median(dat * x * y / np.sqrt(np.abs(x * y) + self.softening**2))
        # mu20 = np.median(dat * x**2)
        # mu02 = np.median(dat * y**2)
        # mu11 = np.median(dat * x * y)
        M = np.array([[mu20, mu11], [mu11, mu02]])
        ones = np.ones(self.n_components)
        if not self.PA.initialized:
            if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
                self.PA.dynamic_value = ones * np.pi / 2
            else:
                self.PA.dynamic_value = (
                    ones * (0.5 * np.arctan2(2 * mu11, mu20 - mu02) - np.pi / 2) % np.pi
                )
        if not self.q.initialized:
            l = np.sort(np.linalg.eigvals(M))
            if np.any(np.iscomplex(l)) or np.any(~np.isfinite(l)):
                l = (0.7, 1.0)
            self.q.dynamic_value = ones * np.clip(np.sqrt(l[0] / l[1]), 0.1, 0.9)

    @forward
    def transform_coordinates(self, x, y, q, PA):
        x, y = super().transform_coordinates(x, y)
        if PA.numel() == 1:
            x, y = func.rotate(-(PA + np.pi / 2), x, y)
            x = x.repeat(q.shape[0], *[1] * x.ndim)
            y = y.repeat(q.shape[0], *[1] * y.ndim)
        else:
            x, y = torch.vmap(lambda pa: func.rotate(-(pa + np.pi / 2), x, y))(PA)
        y = torch.vmap(lambda q, y: y / q)(q, y)
        return x, y

    @forward
    def brightness(self, x, y, flux, sigma, q):
        x, y = self.transform_coordinates(x, y)
        R = self.radius_metric(x, y)
        return torch.sum(
            torch.vmap(
                lambda A, r, sig, _q: (A / torch.sqrt(2 * np.pi * _q * sig**2))
                * torch.exp(-0.5 * (r / sig) ** 2)
            )(flux, R, sigma, q),
            dim=0,
        )



import torch
import numpy as np

sq_2pi = np.sqrt(2 * np.pi)


def gaussian(R, sigma, flux):
    """Gaussian 1d profile function, specifically designed for pytorch
    operations.

    Parameters:
        R: Radii tensor at which to evaluate the sersic function
        sigma: standard deviation of the gaussian in the same units as R
        I0: central surface density
    """
    return (flux / (sq_2pi * sigma)) * torch.exp(-0.5 * torch.pow(R / sigma, 2))



from copy import deepcopy
from typing import Union

import numpy as np
import torch
from astropy.io import fits

__all__ = (
    "centroids_from_segmentation_map",
    "PA_from_segmentation_map",
    "q_from_segmentation_map",
    "windows_from_segmentation_map",
    "scale_windows",
    "filter_windows",
    "transfer_windows",
)


def _select_img(img, hduli):
    if isinstance(img, str):
        if img.endswith(".fits"):
            hdul = fits.open(img)
            img = hdul[hduli].data
        elif img.endswith(".npy"):
            img = np.load(img)
        else:
            raise ValueError(f"unrecognized file type, should be one of: fits, npy\n{img}")
    return img


def centroids_from_segmentation_map(
    seg_map: Union[np.ndarray, str],
    image: "Image",
    sky_level=None,
    hdul_index_seg: int = 0,
    skip_index: tuple = (0,),
):
    """identify centroid centers for all segments in a segmentation map

    For each segment in the map, computes a flux weighted centroid in
    pixel space. A dictionary of pixel centers is produced where the
    keys of the dictionary correspond to the segment id's.

    Parameters:
    ----------
      seg_map (Union[np.ndarray, str]): A segmentation map which gives the object identity for each pixel
      image (Union[np.ndarray, str]): An Image which will be used in the light weighted center of mass calculation
      hdul_index_seg (int): If reading from a fits file this is the hdu list index at which the map is found. Default: 0
      hdul_index_img (int): If reading from a fits file this is the hdu list index at which the image is found. Default: 0
      skip_index (tuple): Lists which identities (if any) in the segmentation map should be ignored. Default (0,)

    Returns:
      centroids (dict): dictionary of centroid positions matched to each segment ID. The centroids are in pixel coordinates
    """

    seg_map = _select_img(seg_map, hdul_index_seg)

    seg_map = seg_map.T
    if sky_level is None:
        sky_level = np.nanmedian(image.data)

    data = image.data.detach().cpu().numpy() - sky_level
    centroids = {}

    II, JJ = np.meshgrid(np.arange(seg_map.shape[0]), np.arange(seg_map.shape[1]), indexing="ij")

    for index in np.unique(seg_map):
        if index is None or index in skip_index:
            continue
        N = seg_map == index
        icentroid = np.sum(II[N] * data[N]) / np.sum(data[N])
        jcentroid = np.sum(JJ[N] * data[N]) / np.sum(data[N])
        xcentroid, ycentroid = image.pixel_to_plane(
            torch.tensor(icentroid, dtype=image.data.dtype, device=image.data.device),
            torch.tensor(jcentroid, dtype=image.data.dtype, device=image.data.device),
            params=(),
        )
        centroids[index] = [xcentroid.item(), ycentroid.item()]

    return centroids


def PA_from_segmentation_map(
    seg_map: Union[np.ndarray, str],
    image: "Image",
    centroids=None,
    sky_level=None,
    hdul_index_seg: int = 0,
    skip_index: tuple = (0,),
    softening=1e-3,
):

    seg_map = _select_img(seg_map, hdul_index_seg)

    # reverse to match numpy indexing
    seg_map = seg_map.T
    if sky_level is None:
        sky_level = np.nanmedian(image.data)

    data = image.data.detach().cpu().numpy() - sky_level

    if centroids is None:
        centroids = centroids_from_segmentation_map(
            seg_map=seg_map, image=image, skip_index=skip_index
        )

    x, y = image.coordinate_center_meshgrid()
    x = x.detach().cpu().numpy()
    y = y.detach().cpu().numpy()
    PAs = {}
    for index in np.unique(seg_map):
        if index is None or index in skip_index:
            continue
        N = seg_map == index
        xx = x[N] - centroids[index][0]
        yy = y[N] - centroids[index][1]
        mu20 = np.median(data[N] * np.abs(xx))
        mu02 = np.median(data[N] * np.abs(yy))
        mu11 = np.median(data[N] * xx * yy / np.sqrt(np.abs(xx * yy) + softening**2))
        M = np.array([[mu20, mu11], [mu11, mu02]])
        if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
            PAs[index] = np.pi / 2
        else:
            PAs[index] = (0.5 * np.arctan2(2 * mu11, mu20 - mu02) - np.pi / 2) % np.pi

    return PAs


def q_from_segmentation_map(
    seg_map: Union[np.ndarray, str],
    image: "Image",
    centroids=None,
    sky_level=None,
    hdul_index_seg: int = 0,
    skip_index: tuple = (0,),
    softening=1e-3,
):

    seg_map = _select_img(seg_map, hdul_index_seg)

    # reverse to match numpy indexing
    seg_map = seg_map.T

    if sky_level is None:
        sky_level = np.nanmedian(image.data)

    data = image.data.detach().cpu().numpy() - sky_level

    if centroids is None:
        centroids = centroids_from_segmentation_map(
            seg_map=seg_map, image=image, skip_index=skip_index
        )

    x, y = image.coordinate_center_meshgrid()
    x = x.detach().cpu().numpy()
    y = y.detach().cpu().numpy()
    qs = {}
    for index in np.unique(seg_map):
        if index is None or index in skip_index:
            continue
        N = seg_map == index
        xx = x[N] - centroids[index][0]
        yy = y[N] - centroids[index][1]
        mu20 = np.median(data[N] * np.abs(xx))
        mu02 = np.median(data[N] * np.abs(yy))
        mu11 = np.median(data[N] * xx * yy / np.sqrt(np.abs(xx * yy) + softening**2))
        M = np.array([[mu20, mu11], [mu11, mu02]])
        if np.any(np.iscomplex(M)) or np.any(~np.isfinite(M)):
            qs[index] = 0.7
        else:
            l = np.abs(np.sort(np.linalg.eigvals(M)))
            qs[index] = np.clip(np.sqrt(l[0] / l[1]), 0.1, 0.9)

    return qs


def windows_from_segmentation_map(seg_map, hdul_index=0, skip_index=(0,)):
    """Convert a segmentation map into boinding boxes

    Takes a segmentation map as input and uses the segmentation ids to
    determine bounding boxes for every object. Scales the bounding
    boxes according to given factors and returns the coordinates.

    each window is formatted as a list of lists with:
    window = [[xmin,ymin],[xmax,ymax]]

    expand_scale changes the base window by the given
    factor. expand_border is added afterwards on all sides (so an
    expand border of 1 will add 2 to the total width of the window.

    """

    if isinstance(seg_map, str):
        if seg_map.endswith(".fits"):
            hdul = fits.open(seg_map)
            seg_map = hdul[hdul_index].data
        elif seg_map.endswith(".npy"):
            seg_map = np.load(seg_map)
        else:
            raise ValueError(f"unrecognized file type, should be one of: fits, npy\n{seg_map}")

    seg_map = seg_map.T

    windows = {}

    for index in np.unique(seg_map):
        if index is None or index in skip_index:
            continue
        Iid, Jid = np.where(seg_map == index)
        # Get window from segmap
        windows[index] = [[np.min(Iid), np.min(Jid)], [np.max(Iid), np.max(Jid)]]

    return windows


def scale_windows(windows, image: "Image" = None, expand_scale=1.0, expand_border=0.0):
    new_windows = {}
    for index in list(windows.keys()):
        new_window = deepcopy(windows[index])
        # Get center and shape of the window
        center = (
            (new_window[0][0] + new_window[1][0]) / 2,
            (new_window[0][1] + new_window[1][1]) / 2,
        )
        shape = (
            new_window[1][0] - new_window[0][0],
            new_window[1][1] - new_window[0][1],
        )
        # Update the window with any expansion coefficients
        new_window = [
            [
                int(center[0] - expand_scale * shape[0] / 2 - expand_border),
                int(center[1] - expand_scale * shape[1] / 2 - expand_border),
            ],
            [
                int(center[0] + expand_scale * shape[0] / 2 + expand_border),
                int(center[1] + expand_scale * shape[1] / 2 + expand_border),
            ],
        ]
        # Ensure the window does not exceed the borders of the image
        if image is not None:
            new_window = [
                [max(0, new_window[0][0]), max(0, new_window[0][1])],
                [
                    min(image.data.shape[0], new_window[1][0]),
                    min(image.data.shape[1], new_window[1][1]),
                ],
            ]
        new_windows[index] = new_window
    return new_windows


def filter_windows(
    windows,
    min_size=None,
    max_size=None,
    min_area=None,
    max_area=None,
    min_flux=None,
    max_flux=None,
    image: "Image" = None,
):
    """
    Filter a set of windows based on a set of criteria.

    Parameters
    ----------
        min_size: minimum size of the window in pixels
        max_size: maximum size of the window in pixels
        min_area: minimum area of the window in pixels
        max_area: maximum area of the window in pixels
        min_flux: minimum flux of the window in ADU
        max_flux: maximum flux of the window in ADU
        image: the image from which the flux is calculated for min_flux and max_flux
    """
    new_windows = {}
    for w in list(windows.keys()):
        if min_size is not None:
            if (
                min(
                    windows[w][1][0] - windows[w][0][0],
                    windows[w][1][1] - windows[w][0][1],
                )
                < min_size
            ):
                continue
        if max_size is not None:
            if (
                max(
                    windows[w][1][0] - windows[w][0][0],
                    windows[w][1][1] - windows[w][0][1],
                )
                > max_size
            ):
                continue
        if min_area is not None:
            if (
                (windows[w][1][0] - windows[w][0][0]) * (windows[w][1][1] - windows[w][0][1])
            ) < min_area:
                continue
        if max_area is not None:
            if (
                (windows[w][1][0] - windows[w][0][0]) * (windows[w][1][1] - windows[w][0][1])
            ) > max_area:
                continue
        if min_flux is not None:
            if (
                np.sum(
                    image.data[
                        windows[w][0][0] : windows[w][1][0],
                        windows[w][0][1] : windows[w][1][1],
                    ]
                )
                < min_flux
            ):
                continue
        if max_flux is not None:
            if (
                np.sum(
                    image.data[
                        windows[w][0][0] : windows[w][1][0],
                        windows[w][0][1] : windows[w][1][1],
                    ]
                )
                > max_flux
            ):
                continue
        new_windows[w] = windows[w]
    return new_windows


def transfer_windows(windows, base_image, new_image):
    """
    Convert a set of windows from one image object to another. This will account
    for the relative adjustments in origin, pixelscale, and rotation between the
    two images.

    Parameters
    ----------
    windows : dict
        A dictionary of windows to be transferred. Each window is formatted as a list of lists with:
        window = [[xmin,ymin],[xmax,ymax]]
    base_image : Image
        The image object from which the windows are being transferred.
    new_image : Image
        The image object to which the windows are being transferred.
    """
    new_windows = {}
    for w in list(windows.keys()):
        four_corners_base = torch.tensor(
            [
                windows[w][0],
                windows[w][1],
                [windows[w][0][0], windows[w][1][1]],
                [windows[w][1][0], windows[w][0][1]],
            ],
            dtype=base_image.data.dtype,
            device=base_image.data.device,
        )  # (4,2)
        four_corners_new = (
            torch.stack(
                new_image.plane_to_pixel(*base_image.pixel_to_plane(*four_corners_base.T)), dim=-1
            )
            .detach()
            .cpu()
            .numpy()
        )  # (4,2)

        bottom_corner = np.floor(np.min(four_corners_new, axis=0)).astype(int)
        bottom_corner = np.clip(bottom_corner, 0, np.array(new_image.shape))
        top_corner = np.ceil(np.max(four_corners_new, axis=0)).astype(int)
        top_corner = np.clip(top_corner, 0, np.array(new_image.shape))
        new_windows[w] = [
            [int(bottom_corner[0]), int(bottom_corner[1])],
            [int(top_corner[0]), int(top_corner[1])],
        ]
    return new_windows



__all__ = ("AstroPhotError", "NameNotAllowed", "SpecificationConflict")


class AstroPhotError(Exception):
    """
    Base exception for all AstroPhot processes.
    """

    ...


class NameNotAllowed(AstroPhotError):
    """
    Used for invalid names of AstroPhot objects
    """

    ...


class SpecificationConflict(AstroPhotError):
    """
    Raised when the inputs to an object are conflicting and/or ambiguous
    """

    ...



import torch

from .psf_model_object import PSFModel
from ..utils.decorators import ignore_numpy_warnings
from ..utils.interpolate import interp2d
from caskade import OverrideParam
from ..param import forward

__all__ = ["PixelatedPSF"]


class PixelatedPSF(PSFModel):
    """point source model which uses an image of the PSF as its
    representation for point sources. Using bilinear interpolation it
    will shift the PSF within a pixel to accurately represent the
    center location of a point source. There is no functional form for
    this object type as any image can be supplied. The image pixels
    will be optimized as individual parameters. This can very quickly
    result in a large number of parameters and a near impossible
    fitting task, ideally this should be restricted to a very small
    area likely at the center of the PSF.

    To initialize the PSF image will by default be set to the target
    PSF_Image values, thus one can use an empirical PSF as a starting
    point. Since only bilinear interpolation is performed, it is
    recommended to provide the PSF at a higher resolution than the
    image if it is near the nyquist sampling limit. Bilinear
    interpolation is very fast and accurate for smooth models, so this
    way it is possible to do the expensive interpolation before
    optimization and save time. Note that if you do this you must
    provide the PSF as a PSF_Image object with the correct pixelscale
    (essentially just divide the pixelscale by the upsampling factor
    you used).

    Parameters:
        pixels: the total flux within each pixel, represented as the log of the flux.

    """

    _model_type = "pixelated"
    _parameter_specs = {"pixels": {"units": "flux/arcsec^2"}}
    usable = True
    sampling_mode = "midpoint"
    integrate_mode = "none"

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()
        if self.pixels.initialized:
            return
        target_area = self.target[self.window]
        self.pixels.dynamic_value = target_area.data.clone() / target_area.pixel_area

    @forward
    def brightness(self, x, y, pixels, center):
        with OverrideParam(self.target.crtan, center):
            pX, pY = self.target.plane_to_pixel(x, y)
        result = interp2d(pixels, pY, pX)
        return result



from typing import Optional

import torch
import numpy as np

from .base import Model
from .model_object import ComponentModel
from ..utils.decorators import ignore_numpy_warnings
from ..utils.interpolate import interp2d
from ..image import Window, PSFImage
from ..errors import SpecificationConflict
from ..param import forward

__all__ = ("PointSource",)


class PointSource(ComponentModel):
    """Describes a point source in the image, this is a delta function at
    some position in the sky. This is typically used to describe
    stars, supernovae, very small galaxies, quasars, asteroids or any
    other object which can essentially be entirely described by a
    position and total flux (no structure).

    """

    _model_type = "point"
    _parameter_specs = {
        "flux": {"units": "flux", "valid": (0, None), "shape": ()},
    }
    usable = True

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        if self.psf is None:
            raise SpecificationConflict("Point_Source needs a psf!")

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.flux.initialized:
            return
        target_area = self.target[self.window]
        dat = target_area.data.detach().cpu().numpy().copy()
        edge = np.concatenate((dat[:, 0], dat[:, -1], dat[0, :], dat[-1, :]))
        edge_average = np.median(edge)
        self.flux.dynamic_value = np.abs(np.sum(dat - edge_average))

    # Psf convolution should be on by default since this is a delta function
    @property
    def psf_mode(self):
        return "full"

    @psf_mode.setter
    def psf_mode(self, value):
        pass

    @property
    def integrate_mode(self):
        return "none"

    @integrate_mode.setter
    def integrate_mode(self, value):
        pass

    @forward
    def sample(self, window: Optional[Window] = None, center=None, flux=None):
        """Evaluate the model on the space covered by an image object. This
        function properly calls integration methods and PSF
        convolution. This should not be overloaded except in special
        cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        PSF convolution, and proper alignment of the computed model
        with the original pixel grid. The final model is then added to
        the requested image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Window within which to evaluate model
        if window is None:
            window = self.window

        if isinstance(self.psf, PSFImage):
            psf = self.psf.data
        elif isinstance(self.psf, Model):
            psf = self.psf().data
        else:
            raise TypeError(
                f"PSF must be a PSFImage or Model instance, got {type(self.psf)} instead."
            )

        # Make the image object to which the samples will be tracked
        working_image = self.target[window].model_image(upsample=self.psf_upscale)

        i, j = working_image.pixel_center_meshgrid()
        i0, j0 = working_image.plane_to_pixel(*center)
        working_image.data = interp2d(
            psf, i - i0 + (psf.shape[0] // 2), j - j0 + (psf.shape[1] // 2)
        )

        working_image.data = flux * working_image.data

        working_image = working_image.reduce(self.psf_upscale)

        return working_image



import astrophot as ap
import torch
import numpy as np
import torch

import astrophot as ap
from utils import make_basic_sersic, make_basic_gaussian_psf


def test_jointmodel_creation():
    np.random.seed(12345)
    shape = (10, 15)
    tar1 = ap.TargetImage(
        name="target1",
        data=np.random.normal(loc=0, scale=1.4, size=shape),
        pixelscale=0.8,
        variance=np.ones(shape) * (1.4**2),
    )
    shape2 = (33, 42)
    tar2 = ap.TargetImage(
        name="target2",
        data=np.random.normal(loc=0, scale=1.4, size=shape2),
        pixelscale=0.3,
        variance=np.ones(shape2) * (1.4**2),
    )

    tar = ap.TargetImageList([tar1, tar2])

    mod1 = ap.models.FlatSky(
        name="base model 1",
        target=tar1,
    )
    mod2 = ap.models.FlatSky(
        name="base model 2",
        target=tar2,
    )

    smod = ap.Model(
        name="group model",
        model_type="group model",
        models=[mod1, mod2],
        target=tar,
    )

    smod.initialize()
    assert torch.all(torch.isfinite(smod().flatten("data"))).item(), "model_image should be real"

    fm = smod.fit_mask()
    for fmi in fm:
        assert torch.sum(fmi).item() == 0, "this fit_mask should not mask any pixels"


def test_psfgroupmodel_creation():
    tar = make_basic_gaussian_psf()

    mod1 = ap.Model(
        name="base model 1",
        model_type="moffat psf model",
        target=tar,
    )

    mod2 = ap.Model(
        name="base model 2",
        model_type="moffat psf model",
        target=tar,
    )

    smod = ap.Model(
        name="group model",
        model_type="psf group model",
        models=[mod1, mod2],
        target=tar,
    )

    smod.initialize()

    assert torch.all(smod().data >= 0), "PSF group sample should be greater than or equal to zero"


def test_joint_multi_band_multi_object():
    target1 = make_basic_sersic(52, 53, name="target1")
    target2 = make_basic_sersic(48, 65, name="target2")
    target3 = make_basic_sersic(60, 49, name="target3")
    target4 = make_basic_sersic(60, 49, name="target4")

    # fmt: off
    model11 = ap.Model(name="model11", model_type="sersic galaxy model", window=(0, 50, 5, 52), target=target1)
    model12 = ap.Model(name="model12", model_type="sersic galaxy model", window=(3, 53, 0, 49), target=target1)
    model1 = ap.Model(name="model1", model_type="group model", models=[model11, model12], target=target1)

    model21 = ap.Model(name="model21", model_type="sersic galaxy model", window=(1, 62, 10, 48), target=target2)
    model22 = ap.Model(name="model22", model_type="sersic galaxy model", window=(2, 60, 5, 49), target=target2)
    model2 = ap.Model(name="model2", model_type="group model", models=[model21, model22], target=target2)

    model31 = ap.Model(name="model31", model_type="sersic galaxy model", window=(1, 62, 10, 48), target=target3)
    model32 = ap.Model(name="model32", model_type="sersic galaxy model", window=(2, 60, 5, 49), target=target3)
    model3 = ap.Model(name="model3", model_type="group model", models=[model31, model32], target=target3)

    model4 = ap.Model(name="model4", model_type="sersic galaxy model", window=(0, 53, 0, 52), target=target1)

    model51 = ap.Model(name="model51", model_type="sersic galaxy model", window=(0, 65, 0, 48), target=target2)
    model52 = ap.Model(name="model52", model_type="sersic galaxy model", window=(0, 49, 0, 60), target=target3)
    model5 = ap.Model(name="model5", model_type="group model", models=[model51, model52], target=ap.TargetImageList([target2, target3]))

    model = ap.Model(name="joint model", model_type="group model", models=[model1, model2, model3, model4, model5], target=ap.TargetImageList([target1, target2, target3, target4]))
    # fmt: on

    model.initialize()
    mask = model.fit_mask()
    assert len(mask) == 4, "There should be 4 fit masks for the 4 targets"
    for m in mask:
        assert torch.all(torch.isfinite(m)), "this fit_mask should be finite"
    sample = model.sample(window=ap.WindowList([target1.window, target2.window, target3.window]))
    assert isinstance(sample, ap.ImageList), "Sample should be an ImageList"
    for image in sample:
        assert torch.all(torch.isfinite(image.data)), "Sample image data should be finite"
        assert torch.all(image.data >= 0), "Sample image data should be non-negative"

    jacobian = model.jacobian()
    assert isinstance(jacobian, ap.ImageList), "Jacobian should be an ImageList"
    for image in jacobian:
        assert torch.all(torch.isfinite(image.data)), "Jacobian image data should be finite"

    window = model.window
    assert isinstance(window, ap.WindowList), "Window should be a WindowList"



from astrophot.param import Param
import torch


def test_param():

    a = Param("a", value=1.0, uncertainty=0.1, valid=(0, 2), prof=1.0)
    assert a.is_valid(1.5), "value should be valid"
    assert isinstance(a.uncertainty, torch.Tensor), "uncertainty should be a tensor"
    assert isinstance(a.prof, torch.Tensor), "prof should be a tensor"
    assert a.initialized, "parameter should be marked as initialized"
    assert a.soft_valid(a.value) == a.value, "soft valid should return the value if not near limits"
    assert (
        a.soft_valid(-1 * torch.ones_like(a.value)) > a.valid[0]
    ), "soft valid should push values inside the limits"
    assert (
        a.soft_valid(3 * torch.ones_like(a.value)) < a.valid[1]
    ), "soft valid should push values inside the limits"

    b = Param("b", value=[2.0, 3.0], uncertainty=[0.1, 0.1], valid=(1, None))
    assert not b.is_valid(0.5), "value should not be valid"
    assert b.is_valid(10.5), "value should be valid"
    assert torch.all(
        b.soft_valid(-1 * torch.ones_like(b.value)) > b.valid[0]
    ), "soft valid should push values inside the limits"
    assert b.prof is None

    c = Param("c", value=lambda P: P.a.value, valid=(None, 4.0))
    c.link(a)
    assert c.initialized, "pointer should be marked as initialized"
    assert c.is_valid(0.5), "value should be valid"
    assert c.uncertainty is None



from .base import AstroPhotError

__all__ = ("InvalidModel", "InvalidTarget", "UnrecognizedModel")


class InvalidModel(AstroPhotError):
    """
    Catches when a model object is inappropriate for this instance.
    """

    ...


class InvalidTarget(AstroPhotError):
    """
    Catches when a target object is assigned incorrectly.
    """

    ...


class UnrecognizedModel(AstroPhotError):
    """
    Raised when the user tries to invoke a model that does not exist.
    """

    ...



import platform
import glob
import pytest
import runpy
import subprocess
import os

pytestmark = pytest.mark.skipif(
    platform.system() in ["Windows", "Darwin"],
    reason="Graphviz not installed on Windows runner",
)

notebooks = glob.glob(
    os.path.join(
        os.path.split(os.path.dirname(__file__))[0], "docs", "source", "tutorials", "*.ipynb"
    )
)


def convert_notebook_to_py(nbpath):
    subprocess.run(
        ["jupyter", "nbconvert", "--to", "python", nbpath],
        check=True,
    )
    pypath = nbpath.replace(".ipynb", ".py")
    with open(pypath, "r") as f:
        content = f.readlines()
    with open(pypath, "w") as f:
        for line in content:
            if line.startswith("get_ipython()"):
                # Remove get_ipython() lines to avoid errors in script execution
                continue
            f.write(line)


def cleanup_py_scripts(nbpath):
    try:
        os.remove(nbpath.replace(".ipynb", ".py"))
        os.remove(nbpath.replace(".ipynb", ".pyc"))
    except FileNotFoundError:
        pass


@pytest.mark.parametrize("nb_path", notebooks)
def test_notebook(nb_path):
    convert_notebook_to_py(nb_path)
    runpy.run_path(nb_path.replace(".ipynb", ".py"), run_name="__main__")
    cleanup_py_scripts(nb_path)



import astrophot as ap
import torch
import numpy as np
from utils import make_basic_gaussian_psf
import pytest

# torch.autograd.set_detect_anomaly(True)
######################################################################
# PSF Model Objects
######################################################################


@pytest.mark.parametrize("model_type", ap.models.PSFModel.List_Models(usable=True, types=True))
def test_all_psfmodel_sample(model_type):

    if "nuker" in model_type:
        kwargs = {"Ib": None}
    elif "gaussian" in model_type:
        kwargs = {"flux": None}
    elif "exponential" in model_type:
        kwargs = {"Ie": None}
    else:
        kwargs = {}
    target = make_basic_gaussian_psf(pixelscale=0.8)
    MODEL = ap.Model(
        name="test model",
        model_type=model_type,
        target=target,
        normalize_psf=False,
        **kwargs,
    )
    MODEL.initialize()
    print(MODEL)
    for P in MODEL.dynamic_params:
        assert P.value is not None, (
            f"Model type {model_type} parameter {P} should not be None after initialization",
        )
    img = MODEL()

    assert torch.all(
        torch.isfinite(img.data)
    ), "Model should evaluate a real number for the full image"

    if model_type == "pixelated psf model":
        psf = ap.utils.initialize.gaussian_psf(3 * 0.8, 25, 0.8)
        MODEL.pixels.dynamic_value = psf / np.sum(psf)

    assert torch.all(
        torch.isfinite(MODEL.jacobian().data)
    ), "Model should evaluate a real number for the jacobian"

    res = ap.fit.LM(MODEL, max_iter=10).fit()

    assert len(res.loss_history) > 2, "Optimizer must be able to find steps to improve the model"

    if "pixelated" in model_type:  # fixme pixelated having difficulties
        return
    assert ((res.loss_history[0] - 1) > (2 * (res.loss_history[-1] - 1))) or (
        res.loss_history[-1] < 1.0
    ), (
        f"Model {model_type} should fit to the target image, but did not. "
        f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
    )



def nuker(R, Rb, Ib, alpha, beta, gamma):
    """Nuker 1d profile function

    Parameters:
        R: Radii tensor at which to evaluate the nuker function
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """
    return (
        Ib
        * (2 ** ((beta - gamma) / alpha))
        * ((R / Rb) ** (-gamma))
        * ((1 + (R / Rb) ** alpha) ** ((gamma - beta) / alpha))
    )



from .base import all_subclasses
from .integration import (
    quad_table,
    pixel_center_integrator,
    pixel_corner_integrator,
    pixel_simpsons_integrator,
    pixel_quad_integrator,
    single_quad_integrate,
    recursive_quad_integrate,
    upsample,
    recursive_bright_integrate,
)
from .convolution import (
    convolve,
    curvature_kernel,
)
from .sersic import sersic, sersic_n_to_b
from .moffat import moffat
from .ferrer import ferrer
from .king import king
from .gaussian import gaussian
from .gaussian_ellipsoid import euler_rotation_matrix
from .exponential import exponential
from .nuker import nuker
from .spline import spline
from .transform import rotate
from .zernike import zernike_n_m_list, zernike_n_m_modes

__all__ = (
    "all_subclasses",
    "quad_table",
    "pixel_center_integrator",
    "pixel_corner_integrator",
    "pixel_simpsons_integrator",
    "pixel_quad_integrator",
    "convolve",
    "curvature_kernel",
    "sersic",
    "sersic_n_to_b",
    "moffat",
    "ferrer",
    "king",
    "gaussian",
    "euler_rotation_matrix",
    "exponential",
    "nuker",
    "spline",
    "single_quad_integrate",
    "recursive_quad_integrate",
    "upsample",
    "recursive_bright_integrate",
    "rotate",
    "zernike_n_m_list",
    "zernike_n_m_modes",
)



import torch

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from ...utils.parametric_profiles import ferrer_np
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from .. import func


def x0_func(model_params, R, F):
    return R[5], 1, 1, 10 ** F[0]


class FerrerMixin:
    """Modified Ferrer radial light profile (Binney & Tremaine 1987).

    This model has a relatively flat brightness core and then a truncation. It
    is used in specialized circumstances such as fitting the bar of a galaxy.
    The functional form of the Modified Ferrer profile is defined as:

    $$I(R) = I_0 \\left(1 - \\left(\\frac{R}{r_{\\rm out}}\\right)^{2-\\beta}\\right)^{\\alpha}$$

    where `rout` is the outer truncation radius, `alpha` controls the steepness
    of the truncation, `beta` controls the shape, and `I0` is the intensity at
    the center of the profile.

    Parameters:
        rout: Outer truncation radius in arcseconds.
        alpha: Inner slope parameter.
        beta: Outer slope parameter.
        I0: Intensity at the center of the profile in flux/arcsec^2
    """

    _model_type = "ferrer"
    _parameter_specs = {
        "rout": {"units": "arcsec", "valid": (0.0, None), "shape": ()},
        "alpha": {"units": "unitless", "valid": (0, 10), "shape": ()},
        "beta": {"units": "unitless", "valid": (0, 2), "shape": ()},
        "I0": {"units": "flux/arcsec^2", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            ferrer_np,
            ("rout", "alpha", "beta", "I0"),
            x0_func,
        )

    @forward
    def radial_model(self, R, rout, alpha, beta, I0):
        return func.ferrer(R, rout, alpha, beta, I0)


class iFerrerMixin:
    """Modified Ferrer radial light profile (Binney & Tremaine 1987).

    This model has a relatively flat brightness core and then a truncation. It
    is used in specialized circumstances such as fitting the bar of a galaxy.
    The functional form of the Modified Ferrer profile is defined as:

    $$I(R) = I_0 \\left(1 - \\left(\\frac{R}{r_{\\rm out}}\\right)^{2-\\beta}\\right)^{\\alpha}$$

    where `rout` is the outer truncation radius, `alpha` controls the steepness
    of the truncation, `beta` controls the shape, and `I0` is the intensity at
    the center of the profile.

    `rout`, `alpha`, `beta`, and `I0` are batched by their first dimension,
    allowing for multiple Ferrer profiles to be defined at once.

    Parameters:
        rout: Outer truncation radius in arcseconds.
        alpha: Inner slope parameter.
        beta: Outer slope parameter.
        I0: Intensity at the center of the profile in flux/arcsec^2
    """

    _model_type = "ferrer"
    _parameter_specs = {
        "rout": {"units": "arcsec", "valid": (0.0, None), "shape": ()},
        "alpha": {"units": "unitless", "valid": (0, 10), "shape": ()},
        "beta": {"units": "unitless", "valid": (0, 2), "shape": ()},
        "I0": {"units": "flux/arcsec^2", "valid": (0.0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=ferrer_np,
            params=("rout", "alpha", "beta", "I0"),
            x0_func=x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, rout, alpha, beta, I0):
        return func.ferrer(R, rout[i], alpha[i], beta[i], I0[i])



import argparse
import requests
import torch
from . import config, models, plots, utils, fit
from .param import forward, Param, Module

from .image import (
    Image,
    ImageList,
    TargetImage,
    TargetImageList,
    SIPTargetImage,
    JacobianImage,
    JacobianImageList,
    PSFImage,
    ModelImage,
    ModelImageList,
    Window,
    WindowList,
)
from .models import Model

try:
    from ._version import version as VERSION  # noqa
except ModuleNotFoundError:
    VERSION = "0.0.0"
    print(
        "WARNING: AstroPhot version number not found. This is likely because you are running AstroPhot from a source directory."
    )


# meta data
__version__ = VERSION
__author__ = "Connor Stone"
__email__ = "connorstone628@gmail.com"


def run_from_terminal() -> None:
    """
    Running from terminal no longer supported. This is only used for convenience to download the tutorials.

    """
    config.logger.debug("running from the terminal, not sure if it will catch me.")
    parser = argparse.ArgumentParser(
        prog="astrophot",
        description="Fast and flexible astronomical image photometry package. For the documentation go to: https://astrophot.readthedocs.io",
        epilog="Please see the documentation or contact connor stone (connorstone628@gmail.com) for further assistance.",
    )
    parser.add_argument(
        "filename",
        nargs="?",
        metavar="configfile",
        help="the path to the configuration file. Or just 'tutorial' to download tutorials.",
    )
    # parser.add_argument(
    #     "--config",
    #     type=str,
    #     default="astrophot",
    #     choices=["astrophot", "galfit"],
    #     metavar="format",
    #     help="The type of configuration file being being provided. One of: astrophot, galfit.",
    # )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="print the current AstroPhot version to screen",
    )
    # parser.add_argument(
    #     "--log",
    #     type=str,
    #     metavar="logfile.log",
    #     help="set the log file name for AstroPhot. use 'none' to suppress the log file.",
    # )
    # parser.add_argument(
    #     "-q",
    #     action="store_true",
    #     help="quiet flag to stop command line output, only print to log file",
    # )
    # parser.add_argument(
    #     "--dtype",
    #     type=str,
    #     choices=["float64", "float32"],
    #     metavar="datatype",
    #     help="set the float point precision. Must be one of: float64, float32",
    # )
    # parser.add_argument(
    #     "--device",
    #     type=str,
    #     choices=["cpu", "gpu"],
    #     metavar="device",
    #     help="set the device for AstroPhot to use for computations. Must be one of: cpu, gpu",
    # )

    args = parser.parse_args()

    if args.log is not None:
        config.set_logging_output(
            stdout=not args.q, filename=None if args.log == "none" else args.log
        )
    elif args.q:
        config.set_logging_output(stdout=not args.q, filename="AstroPhot.log")

    if args.dtype is not None:
        config.DTYPE = torch.float64 if args.dtype == "float64" else torch.float32
    if args.device is not None:
        config.DEVICE = "cpu" if args.device == "cpu" else "cuda:0"

    if args.filename is None:
        raise RuntimeError(
            "Please pass a config file to astrophot. See 'astrophot --help' for more information, or go to https://astrophot.readthedocs.io"
        )
    if args.filename in ["tutorial", "tutorials"]:
        tutorials = [
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/GettingStarted.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/GroupModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/ModelZoo.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/JointModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/FittingMethods.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/CustomModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/BasicPSFModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/AdvancedPSFModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/source/tutorials/ConstrainedModels.ipynb",
        ]
        for url in tutorials:
            try:
                R = requests.get(url)
                with open(url[url.rfind("/") + 1 :], "w") as f:
                    f.write(R.text)
            except:
                print(
                    f"WARNING: couldn't find tutorial: {url[url.rfind('/')+1:]} check internet connection"
                )

        config.logger.info("collected the tutorials")
    else:
        raise ValueError(f"Unrecognized request")


__all__ = (
    "models",
    "Model",
    "Image",
    "ImageList",
    "TargetImage",
    "TargetImageList",
    "SIPTargetImage",
    "JacobianImage",
    "JacobianImageList",
    "PSFImage",
    "ModelImage",
    "ModelImageList",
    "Window",
    "WindowList",
    "plots",
    "utils",
    "fit",
    "forward",
    "Param",
    "Module",
    "config",
    "run_from_terminal",
    "__version__",
    "__author__",
    "__email__",
)



from typing import Optional

import numpy as np
import torch

from ..param import forward
from .base import Model
from . import func
from ..image import (
    TargetImage,
    Window,
    PSFImage,
)
from ..utils.initialize import recursive_center_of_mass
from ..utils.decorators import ignore_numpy_warnings
from .. import config
from ..errors import InvalidTarget
from .mixins import SampleMixin

__all__ = ["ComponentModel"]


class ComponentModel(SampleMixin, Model):
    """Component of a model for an object in an image.

    This is a single component of an image model. It has a position on the sky
    determined by `center` and may or may not be convolved with a PSF to represent some data.

    Options:
        psf_convolve: Whether to convolve the model with a PSF. (bool)

    """

    _parameter_specs = {"center": {"units": "arcsec", "shape": (2,)}}

    _options = ("psf_convolve",)
    psf_convolve: bool = False

    usable = False

    def __init__(self, *args, psf=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.psf = psf

    @property
    def psf(self):
        if self._psf is None:
            return self.target.psf
        return self._psf

    @psf.setter
    def psf(self, val):
        try:
            del self._psf  # Remove old PSF if it exists
        except AttributeError:
            pass
        if val is None:
            self._psf = None
        elif isinstance(val, PSFImage):
            self._psf = val
            self.psf_convolve = True
        elif isinstance(val, Model):
            self._psf = val
            self.psf_convolve = True
        else:
            self._psf = self.target.psf_image(data=val)
            self.psf_convolve = True
        self.update_psf_upscale()

    def update_psf_upscale(self):
        """Update the PSF upscale factor based on the current target pixel length."""
        if self.psf is None:
            self.psf_upscale = 1
        elif isinstance(self.psf, PSFImage):
            self.psf_upscale = (
                torch.round(self.target.pixelscale / self.psf.pixelscale).int().item()
            )
        elif isinstance(self.psf, Model):
            self.psf_upscale = (
                torch.round(self.target.pixelscale / self.psf.target.pixelscale).int().item()
            )
        else:
            raise TypeError(
                f"PSF must be a PSFImage or Model instance, got {type(self.psf)} instead."
            )

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, tar):
        if tar is None:
            self._target = None
            return
        elif not isinstance(tar, TargetImage):
            raise InvalidTarget("AstroPhot Model target must be a TargetImage instance.")
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass
        self._target = tar
        try:
            self.update_psf_upscale()
        except AttributeError:
            pass

    # Initialization functions
    ######################################################################
    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        """Determine initial values for the center coordinates. This is done
        with a local center of mass search which iterates by finding
        the center of light in a window, then iteratively updates
        until the iterations move by less than a pixel.

        Args:
          target (Optional[Target_Image]): A target image object to use as a reference when setting parameter values

        """
        if self.psf is not None and isinstance(self.psf, Model):
            self.psf.initialize()

        # Use center of window if a center hasn't been set yet
        if self.center.initialized:
            return

        target_area = self.target[self.window]
        dat = np.copy(target_area.data.detach().cpu().numpy())
        if target_area.has_mask:
            mask = target_area.mask.detach().cpu().numpy()
            dat[mask] = np.nanmedian(dat[~mask])

        COM = recursive_center_of_mass(dat)
        if not np.all(np.isfinite(COM)):
            return
        COM_center = target_area.pixel_to_plane(
            *torch.tensor(COM, dtype=config.DTYPE, device=config.DEVICE)
        )
        self.center.dynamic_value = COM_center

    def fit_mask(self):
        return torch.zeros_like(self.target[self.window].mask, dtype=torch.bool)

    @forward
    def transform_coordinates(self, x, y, center):
        return x - center[0], y - center[1]

    @forward
    def sample(
        self,
        window: Optional[Window] = None,
    ):
        """Evaluate the model on the pixels defined in an image. This
        function properly calls integration methods and PSF
        convolution. This should not be overloaded except in special
        cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        PSF convolution, and proper alignment of the computed model
        with the original pixel grid. The final model is then added to
        the requested image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Window within which to evaluate model
        if window is None:
            window = self.window

        if self.psf_convolve:
            psf = self.psf() if isinstance(self.psf, Model) else self.psf

            working_image = self.target[window].model_image(
                upsample=self.psf_upscale, pad=psf.psf_pad
            )
            sample = self.sample_image(working_image)
            working_image._data = func.convolve(sample, psf.data)
            working_image = working_image.crop(psf.psf_pad).reduce(self.psf_upscale)

        else:
            working_image = self.target[window].model_image()
            working_image._data = self.sample_image(working_image)

        # Units from flux/arcsec^2 to flux, multiply by pixel area
        working_image.fluxdensity_to_flux()

        return working_image



from .brightness import RadialMixin, WedgeMixin, RayMixin
from .transform import (
    InclinedMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    TruncationMixin,
)
from .sersic import SersicMixin, iSersicMixin
from .exponential import ExponentialMixin, iExponentialMixin
from .moffat import MoffatMixin, iMoffatMixin
from .ferrer import FerrerMixin, iFerrerMixin
from .king import KingMixin, iKingMixin
from .gaussian import GaussianMixin, iGaussianMixin
from .nuker import NukerMixin, iNukerMixin
from .spline import SplineMixin, iSplineMixin
from .sample import SampleMixin

__all__ = (
    "RadialMixin",
    "WedgeMixin",
    "RayMixin",
    "SuperEllipseMixin",
    "FourierEllipseMixin",
    "WarpMixin",
    "TruncationMixin",
    "InclinedMixin",
    "SersicMixin",
    "iSersicMixin",
    "ExponentialMixin",
    "iExponentialMixin",
    "MoffatMixin",
    "iMoffatMixin",
    "FerrerMixin",
    "iFerrerMixin",
    "KingMixin",
    "iKingMixin",
    "GaussianMixin",
    "iGaussianMixin",
    "NukerMixin",
    "iNukerMixin",
    "SplineMixin",
    "iSplineMixin",
    "SampleMixin",
)



import sys
import logging
import torch

__all__ = ["DTYPE", "DEVICE", "logger", "set_logging_output"]

DTYPE = torch.float64
DEVICE = "cuda:0" if torch.cuda.is_available() else "cpu"

logging.basicConfig(
    filename="AstroPhot.log",
    level=logging.INFO,
    format="%(asctime)s:%(levelname)s: %(message)s",
)
logger = logging.getLogger()
out_handler = logging.StreamHandler(sys.stdout)
out_handler.setLevel(logging.INFO)
out_handler.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(out_handler)


def set_logging_output(stdout=True, filename=None, **kwargs):
    """
    Change the logging system for AstroPhot.
    Here you can set whether output prints to screen or to a logging file.
    This function will remove all handlers from the current logger in logger,
    then add new handlers based on the input to the function.

    Parameters:
        stdout (bool): If True, log messages will be printed to the console (standard output). Default is True.
        filename (str): If given as a string, this will be the name of the file that log messages are written to.
                        If None, no logging file will be used. Default is None.
        stdout_level (logging.INFO): The logging level of messages written to stdout, this can be different from the file level.
                            Default is logging.INFO.
        stdout_formatter (logging.Formatter): A logging.Formatter object which determines what information to include with the logging message only when printing to stdout. Default is logging.Formatter('%(message)s').
        filename_level (int): The logging level of messages written to the log file, this can be different from the stdout level. Default is logging.INFO.
        filename_formatter (logging.Formatter): A logging.Formatter object which determines what information to include with the logging message only when printing to the log file. Default is logging.Formatter('%(asctime)s:%(levelname)s: %(message)s').

    """
    hi = 0
    while hi < len(logger.handlers):
        if isinstance(logger.handlers[hi], logging.StreamHandler):
            logger.removeHandler(logger.handlers[hi])
        elif isinstance(logger.handlers[hi], logging.FileHandler):
            logger.removeHandler(logger.handlers[hi])
        else:
            hi += 1

    if stdout:
        out_handler = logging.StreamHandler(sys.stdout)
        out_handler.setLevel(kwargs.get("stdout_level", logging.INFO))
        out_handler.setFormatter(kwargs.get("stdout_formatter", logging.Formatter("%(message)s")))
        logger.addHandler(out_handler)
        logger.debug("logging now going to stdout")
    if filename is not None:
        out_handler = logging.FileHandler(filename)
        out_handler.setLevel(kwargs.get("filename_level", logging.INFO))
        out_handler.setFormatter(
            kwargs.get(
                "filename_formatter",
                logging.Formatter("%(asctime)s:%(levelname)s: %(message)s"),
            )
        )
        logger.addHandler(out_handler)
        logger.debug("logging now going to %s" % filename)



# Base model object
from .base import Model

# Primary model types
from .model_object import ComponentModel
from .psf_model_object import PSFModel
from .group_model_object import GroupModel
from .group_psf_model import PSFGroupModel

# Component model main types
from .galaxy_model_object import GalaxyModel
from .sky_model_object import SkyModel
from .point_source import PointSource

# subtypes of PSFModel
from .basis import PixelBasisPSF
from .airy import AiryPSF
from .pixelated_psf import PixelatedPSF

# Subtypes of SkyModel
from .flatsky import FlatSky
from .planesky import PlaneSky
from .bilinear_sky import BilinearSky

# Special galaxy types
from .edgeon import EdgeonModel, EdgeonSech, EdgeonIsothermal
from .multi_gaussian_expansion import MultiGaussianExpansion
from .gaussian_ellipsoid import GaussianEllipsoid

# Standard models based on a core radial profile
from .sersic import (
    SersicGalaxy,
    SersicPSF,
    SersicFourierEllipse,
    SersicSuperEllipse,
    SersicWarp,
    SersicRay,
    SersicWedge,
)
from .exponential import (
    ExponentialGalaxy,
    ExponentialPSF,
    ExponentialSuperEllipse,
    ExponentialFourierEllipse,
    ExponentialWarp,
    ExponentialRay,
    ExponentialWedge,
)
from .gaussian import (
    GaussianGalaxy,
    GaussianPSF,
    GaussianSuperEllipse,
    GaussianFourierEllipse,
    GaussianWarp,
    GaussianRay,
    GaussianWedge,
)
from .moffat import (
    MoffatGalaxy,
    MoffatPSF,
    Moffat2DPSF,
    MoffatFourierEllipse,
    MoffatRay,
    MoffatWedge,
    MoffatWarp,
    MoffatSuperEllipse,
)
from .ferrer import (
    FerrerGalaxy,
    FerrerPSF,
    FerrerSuperEllipse,
    FerrerFourierEllipse,
    FerrerWarp,
    FerrerRay,
    FerrerWedge,
)
from .king import (
    KingGalaxy,
    KingPSF,
    KingSuperEllipse,
    KingFourierEllipse,
    KingWarp,
    KingRay,
    KingWedge,
)
from .nuker import (
    NukerGalaxy,
    NukerPSF,
    NukerFourierEllipse,
    NukerSuperEllipse,
    NukerWarp,
    NukerRay,
    NukerWedge,
)
from .spline import (
    SplineGalaxy,
    SplinePSF,
    SplineFourierEllipse,
    SplineSuperEllipse,
    SplineWarp,
    SplineRay,
    SplineWedge,
)

from .mixins import (
    RadialMixin,
    WedgeMixin,
    RayMixin,
    ExponentialMixin,
    iExponentialMixin,
    FerrerMixin,
    iFerrerMixin,
    GaussianMixin,
    iGaussianMixin,
    KingMixin,
    iKingMixin,
    MoffatMixin,
    iMoffatMixin,
    NukerMixin,
    iNukerMixin,
    SersicMixin,
    iSersicMixin,
    SplineMixin,
    iSplineMixin,
    SampleMixin,
    InclinedMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
    TruncationMixin,
)


__all__ = (
    "Model",
    "ComponentModel",
    "PSFModel",
    "GroupModel",
    "PSFGroupModel",
    "GalaxyModel",
    "SkyModel",
    "PointSource",
    "RayGalaxy",
    "SuperEllipseGalaxy",
    "WedgeGalaxy",
    "WarpGalaxy",
    "PixelBasisPSF",
    "AiryPSF",
    "PixelatedPSF",
    "FlatSky",
    "PlaneSky",
    "BilinearSky",
    "EdgeonModel",
    "EdgeonSech",
    "EdgeonIsothermal",
    "MultiGaussianExpansion",
    "GaussianEllipsoid",
    "FourierEllipseGalaxy",
    "SersicGalaxy",
    "SersicPSF",
    "SersicFourierEllipse",
    "SersicSuperEllipse",
    "SersicWarp",
    "SersicRay",
    "SersicWedge",
    "ExponentialGalaxy",
    "ExponentialPSF",
    "ExponentialSuperEllipse",
    "ExponentialFourierEllipse",
    "ExponentialWarp",
    "ExponentialRay",
    "ExponentialWedge",
    "GaussianGalaxy",
    "GaussianPSF",
    "GaussianSuperEllipse",
    "GaussianFourierEllipse",
    "GaussianWarp",
    "GaussianRay",
    "GaussianWedge",
    "MoffatGalaxy",
    "MoffatPSF",
    "Moffat2DPSF",
    "MoffatFourierEllipse",
    "MoffatRay",
    "MoffatWedge",
    "MoffatWarp",
    "MoffatSuperEllipse",
    "FerrerGalaxy",
    "FerrerPSF",
    "FerrerSuperEllipse",
    "FerrerFourierEllipse",
    "FerrerWarp",
    "FerrerRay",
    "FerrerWedge",
    "KingGalaxy",
    "KingPSF",
    "KingSuperEllipse",
    "KingFourierEllipse",
    "KingWarp",
    "KingRay",
    "KingWedge",
    "NukerGalaxy",
    "NukerPSF",
    "NukerFourierEllipse",
    "NukerSuperEllipse",
    "NukerWarp",
    "NukerRay",
    "NukerWedge",
    "SplineGalaxy",
    "SplinePSF",
    "SplineFourierEllipse",
    "SplineWarp",
    "SplineSuperEllipse",
    "SplineRay",
    "SplineWedge",
    "RadialMixin",
    "WedgeMixin",
    "RayMixin",
    "ExponentialMixin",
    "iExponentialMixin",
    "FerrerMixin",
    "iFerrerMixin",
    "GaussianMixin",
    "iGaussianMixin",
    "KingMixin",
    "iKingMixin",
    "MoffatMixin",
    "iMoffatMixin",
    "NukerMixin",
    "iNukerMixin",
    "SersicMixin",
    "iSersicMixin",
    "SplineMixin",
    "iSplineMixin",
    "SampleMixin",
    "InclinedMixin",
    "SuperEllipseMixin",
    "FourierEllipseMixin",
    "WarpMixin",
    "TruncationMixin",
)



import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.stats import binned_statistic
import torch
from ...errors import InvalidData
import matplotlib.pyplot as plt


def auto_variance(data, mask=None):

    if isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()
    if isinstance(mask, torch.Tensor):
        mask = mask.detach().cpu().numpy()
    if mask is None:
        mask = np.zeros(data.shape, dtype=int)

    # Data too small for anything fancy
    var = np.var(data[np.logical_not(mask)])
    if not np.isfinite(var) or var == 0:
        return np.ones_like(data)
    if min(data.shape) < 20:
        return np.ones_like(data) * var

    # Get approximation of noise in each pixel
    blur_data = gaussian_filter(data, 1.1)[4:-4, 4:-4]
    blur_mask = gaussian_filter(mask, 1.1)[4:-4, 4:-4] == 0
    mask = np.logical_not(mask)
    residuals = (data[4:-4, 4:-4] - blur_data)[blur_mask]

    # Bin the residuals by flux, clip the tails
    clips = (np.quantile(data[mask], 0.01), np.quantile(data[mask], 0.99))
    bins = np.linspace(clips[0], clips[1], 11)
    std, bins, _ = binned_statistic(
        data[4:-4, 4:-4][blur_mask].flatten(),
        residuals.flatten(),
        statistic="std",
        bins=bins,
    )
    N = np.logical_not(np.isfinite(std))
    if np.any(N):
        std[N] = np.sqrt(np.interp(bins[:-1][N], bins[:-1][~N], std[~N] ** 2))

    # Fit a line to the variance
    p = np.polyfit(bins[:-3], std[:-2] ** 2, 1)

    # Check if the variance is increasing with flux
    if p[0] < 0:
        return np.ones_like(data) * var
    # Compute the approximate variance map
    variance = np.clip(p[0] * data + p[1], np.min(std) ** 2, None)
    variance[np.logical_not(mask)] = np.inf
    return variance



import numpy as np


def gaussian_psf(sigma, img_width, pixelscale, upsample=4, normalize=True):
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Gaussian at each pixel
    ZZ = np.exp(-0.5 * (XX**2 + YY**2) / sigma**2)

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (upsample**2)

    # Normalize the PSF
    if normalize:
        return ZZ / np.sum(ZZ)
    return ZZ


def moffat_psf(n, Rd, img_width, pixelscale, upsample=4, normalize=True):
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Moffat at each pixel
    ZZ = 1.0 / (1.0 + (XX**2 + YY**2) / (Rd**2)) ** n

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (upsample**2)

    # Normalize the PSF
    if normalize:
        return ZZ / np.sum(ZZ)
    return ZZ



from typing import Optional, Union
from copy import deepcopy

import torch
from torch.func import hessian
import numpy as np

from caskade import Param as CParam
from ..param import Module, forward, Param
from ..utils.decorators import classproperty
from ..image import Window, ImageList, ModelImage, ModelImageList
from ..errors import UnrecognizedModel, InvalidWindow
from .. import config
from . import func

__all__ = ("Model",)


######################################################################
class Model(Module):
    """Base class for all AstroPhot models."""

    _model_type = "model"
    _parameter_specs = {}
    # Softening length used for numerical stability and/or integration stability to avoid discontinuities (near R=0)
    softening = 1e-3  # arcsec
    _options = ("softening",)
    usable = False

    def __new__(cls, *, filename=None, model_type=None, **kwargs):
        if filename is not None:
            state = Model.load(filename)
            MODELS = Model.List_Models()
            for M in MODELS:
                if M.model_type == state["model_type"]:
                    return super(Model, cls).__new__(M)
            else:
                raise UnrecognizedModel(f"Unknown AstroPhot model type: {state['model_type']}")
        elif model_type is not None:
            MODELS = Model.List_Models()  # all_subclasses(Model)
            for M in MODELS:
                if M.model_type == model_type:
                    return super(Model, cls).__new__(M)
            else:
                raise UnrecognizedModel(f"Unknown AstroPhot model type: {model_type}")

        return super().__new__(cls)

    def __init__(self, *, name=None, target=None, window=None, mask=None, filename=None, **kwargs):
        super().__init__(name=name)
        self.target = target
        self.window = window
        self.mask = mask

        # Set any user defined options for the model
        for kwarg in list(kwargs.keys()):
            if kwarg in self.options:
                setattr(self, kwarg, kwargs.pop(kwarg))

        # Create Param objects for this Module
        parameter_specs = self.build_parameter_specs(kwargs, self.parameter_specs)
        for key in parameter_specs:
            param = Param(key, **parameter_specs[key], dtype=config.DTYPE, device=config.DEVICE)
            setattr(self, key, param)

        self.saveattrs.update(self.options)
        self.saveattrs.add("window.extent")

        kwargs.pop("model_type", None)  # model_type is set by __new__
        if len(kwargs) > 0:
            raise TypeError(
                f"Unrecognized keyword arguments for {self.__class__.__name__}: {', '.join(kwargs.keys())}"
            )

    @classproperty
    def model_type(cls) -> str:
        collected = []
        for subcls in cls.mro():
            if subcls is object:
                continue
            mt = subcls.__dict__.get("_model_type", None)
            if mt:
                collected.append(mt)
        return " ".join(collected)

    @classproperty
    def options(cls) -> set:
        options = set()
        for subcls in cls.mro():
            if subcls is object:
                continue
            options.update(subcls.__dict__.get("_options", []))
        return options

    @classproperty
    def parameter_specs(cls) -> dict:
        """Collects all parameter specifications from the class hierarchy."""
        specs = {}
        for subcls in reversed(cls.mro()):
            if subcls is object:
                continue
            specs.update(getattr(subcls, "_parameter_specs", {}))
        return specs

    def build_parameter_specs(self, kwargs, parameter_specs) -> dict:
        parameter_specs = deepcopy(parameter_specs)

        for p in list(kwargs.keys()):
            if p not in parameter_specs:
                continue
            if isinstance(kwargs[p], dict):
                parameter_specs[p].update(kwargs.pop(p))
            else:
                parameter_specs[p]["dynamic_value"] = kwargs.pop(p)
                parameter_specs[p].pop("value", None)
            if isinstance(parameter_specs[p].get("dynamic_value", None), CParam) or callable(
                parameter_specs[p].get("dynamic_value", None)
            ):
                parameter_specs[p]["value"] = parameter_specs[p]["dynamic_value"]
                parameter_specs[p].pop("dynamic_value", None)

        return parameter_specs

    @forward
    def gaussian_log_likelihood(
        self,
        window: Optional[Window] = None,
    ) -> torch.Tensor:
        """
        Compute the negative log likelihood of the model wrt the target image in the appropriate window.
        """

        if window is None:
            window = self.window
        model = self(window=window).data
        data = self.target[window]
        weight = data.weight
        mask = data.mask
        data = data.data
        if isinstance(data, tuple):
            nll = 0.5 * sum(
                torch.sum(((da - mo) ** 2 * wgt)[~ma])
                for mo, da, wgt, ma in zip(model, data, weight, mask)
            )
        else:
            nll = 0.5 * torch.sum(((data - model) ** 2 * weight)[~mask])

        return -nll

    @forward
    def poisson_log_likelihood(
        self,
        window: Optional[Window] = None,
    ) -> torch.Tensor:
        """
        Compute the negative log likelihood of the model wrt the target image in the appropriate window.
        """
        if window is None:
            window = self.window
        model = self(window=window).data
        data = self.target[window]
        mask = data.mask
        data = data.data

        if isinstance(data, tuple):
            nll = sum(
                torch.sum((mo - da * (mo + 1e-10).log() + torch.lgamma(da + 1))[~ma])
                for mo, da, ma in zip(model, data, mask)
            )
        else:
            nll = torch.sum((model - data * (model + 1e-10).log() + torch.lgamma(data + 1))[~mask])

        return -nll

    def hessian(self, likelihood="gaussian"):
        if likelihood == "gaussian":
            return hessian(self.gaussian_log_likelihood)(self.build_params_array())
        elif likelihood == "poisson":
            return hessian(self.poisson_log_likelihood)(self.build_params_array())
        else:
            raise ValueError(f"Unknown likelihood type: {likelihood}")

    def total_flux(self, window=None) -> torch.Tensor:
        F = self(window=window)
        return torch.sum(F.data)

    def total_flux_uncertainty(self, window=None) -> torch.Tensor:
        jac = self.jacobian(window=window).flatten("data")
        dF = torch.sum(jac, dim=0)  # VJP for sum(total_flux)
        current_uncertainty = self.build_params_array_uncertainty()
        return torch.sqrt(torch.sum((dF * current_uncertainty) ** 2))

    def total_magnitude(self, window=None) -> torch.Tensor:
        """Compute the total magnitude of the model in the given window."""
        F = self.total_flux(window=window)
        return -2.5 * torch.log10(F) + self.target.zeropoint

    def total_magnitude_uncertainty(self, window=None) -> torch.Tensor:
        """Compute the uncertainty in the total magnitude of the model in the given window."""
        F = self.total_flux(window=window)
        dF = self.total_flux_uncertainty(window=window)
        return 2.5 * (dF / F) / np.log(10)

    @property
    def window(self) -> Optional[Window]:
        """The window defines a region on the sky in which this model will be
        optimized and typically evaluated. Two models with
        non-overlapping windows are in effect independent of each
        other. If there is another model with a window that spans both
        of them, then they are tenuously connected.

        If not provided, the model will assume a window equal to the
        target it is fitting. Note that in this case the window is not
        explicitly set to the target window, so if the model is moved
        to another target then the fitting window will also change.

        """
        if self._window is None:
            if self.target is None:
                raise ValueError(
                    "This model has no target or window, these must be provided by the user"
                )
            return self.target.window
        return self._window

    @window.setter
    def window(self, window):
        if window is None:
            self._window = None
        elif isinstance(window, Window):
            self._window = window
        elif len(window) in [2, 4]:
            self._window = Window(window, image=self.target)
        else:
            raise InvalidWindow(f"Unrecognized window format: {str(window)}")

    @classmethod
    def List_Models(cls, usable: Optional[bool] = None, types: bool = False) -> set:
        MODELS = func.all_subclasses(cls)
        result = set()
        for model in MODELS:
            if not (model.__dict__.get("usable", False) is usable or usable is None):
                continue
            if types:
                result.add(model.model_type)
            else:
                result.add(model)
        return result

    @forward
    def radius_metric(self, x, y):
        return (x**2 + y**2 + self.softening**2).sqrt()

    @forward
    def angular_metric(self, x, y):
        return torch.atan2(y, x)

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = config.DTYPE
        if device is None:
            device = config.DEVICE
        super().to(dtype=dtype, device=device)

    @forward
    def __call__(
        self,
        window: Optional[Window] = None,
        **kwargs,
    ) -> Union[ModelImage, ModelImageList]:

        return self.sample(window=window, **kwargs)



import torch

from ...utils.integration import quad_table


def pixel_center_meshgrid(shape, dtype, device):
    i = torch.arange(shape[0], dtype=dtype, device=device)
    j = torch.arange(shape[1], dtype=dtype, device=device)
    return torch.meshgrid(i, j, indexing="ij")


def pixel_corner_meshgrid(shape, dtype, device):
    i = torch.arange(shape[0] + 1, dtype=dtype, device=device) - 0.5
    j = torch.arange(shape[1] + 1, dtype=dtype, device=device) - 0.5
    return torch.meshgrid(i, j, indexing="ij")


def pixel_simpsons_meshgrid(shape, dtype, device):
    i = 0.5 * torch.arange(2 * shape[0] + 1, dtype=dtype, device=device) - 0.5
    j = 0.5 * torch.arange(2 * shape[1] + 1, dtype=dtype, device=device) - 0.5
    return torch.meshgrid(i, j, indexing="ij")


def pixel_quad_meshgrid(shape, dtype, device, order=3):
    i, j = pixel_center_meshgrid(shape, dtype, device)
    di, dj, w = quad_table(order, dtype, device)
    i = torch.repeat_interleave(i[..., None], order**2, -1) + di.flatten()
    j = torch.repeat_interleave(j[..., None], order**2, -1) + dj.flatten()
    return i, j, w.flatten()


def rotate(theta, x, y):
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = theta.sin()
    c = theta.cos()
    return c * x - s * y, s * x + c * y



import torch

from ...param import forward
from ...utils.decorators import ignore_numpy_warnings
from .._shared_methods import parametric_initialize, parametric_segment_initialize
from ...utils.parametric_profiles import gaussian_np
from .. import func


def _x0_func(model_params, R, F):
    return R[4], 10 ** F[0]


class GaussianMixin:
    """Gaussian radial light profile.

    The Gaussian profile is a simple and widely used model for extended objects.
    The functional form of the Gaussian profile is defined as:

    $$I(R) = \\frac{{\\rm flux}}{\\sqrt{2\\pi}\\sigma} \exp(-R^2 / (2 \sigma^2))$$

    where `I_0` is the intensity at the center of the profile and `sigma` is the
    standard deviation which controls the width of the profile.

    Parameters:
        sigma: Standard deviation of the Gaussian profile in arcseconds.
        flux: Total flux of the Gaussian profile.
    """

    _model_type = "gaussian"
    _parameter_specs = {
        "sigma": {"units": "arcsec", "valid": (0, None), "shape": ()},
        "flux": {"units": "flux", "valid": (0, None), "shape": ()},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_initialize(
            self,
            self.target[self.window],
            gaussian_np,
            ("sigma", "flux"),
            _x0_func,
        )

    @forward
    def radial_model(self, R, sigma, flux):
        return func.gaussian(R, sigma, flux)


class iGaussianMixin:
    """Gaussian radial light profile.

    The Gaussian profile is a simple and widely used model for extended objects.
    The functional form of the Gaussian profile is defined as:

    $$I(R) = \\frac{{\\rm flux}}{\\sqrt{2\\pi}\\sigma} \exp(-R^2 / (2 \sigma^2))$$

    where `sigma` is the standard deviation which controls the width of the
    profile and `flux` gives the total flux of the profile (assuming no
    perturbations).

    `sigma` and `flux` are batched by their first dimension, allowing for
    multiple Gaussian profiles to be defined at once.

    Parameters:
        sigma: Standard deviation of the Gaussian profile in arcseconds.
        flux: Total flux of the Gaussian profile.
    """

    _model_type = "gaussian"
    _parameter_specs = {
        "sigma": {"units": "arcsec", "valid": (0, None)},
        "flux": {"units": "flux", "valid": (0, None)},
    }

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        parametric_segment_initialize(
            model=self,
            target=self.target[self.window],
            prof_func=gaussian_np,
            params=("sigma", "flux"),
            x0_func=_x0_func,
            segments=self.segments,
        )

    @forward
    def iradial_model(self, i, R, sigma, flux):
        return func.gaussian(R, sigma[i], flux[i])



import numpy as np
import torch

from .sky_model_object import SkyModel
from ..utils.decorators import ignore_numpy_warnings
from ..param import forward

__all__ = ["PlaneSky"]


class PlaneSky(SkyModel):
    """Sky background model using a tilted plane for the sky flux. The brightness for each pixel is defined as:

    I(X, Y) = S + X*dx + Y*dy

    where I(X,Y) is the brightness as a function of image position X Y,
    S is the central sky brightness value, and dx dy are the slopes of
    the sky brightness plane.

    Parameters:
        sky: central sky brightness value
        delta: Tensor for slope of the sky brightness in each image dimension

    """

    _model_type = "plane"
    _parameter_specs = {
        "I0": {"units": "flux/arcsec^2"},
        "delta": {"units": "flux/arcsec"},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if not self.I0.initialized:
            dat = self.target[self.window].data.detach().cpu().numpy().copy()
            self.I0.dynamic_value = np.median(dat) / self.target.pixel_area.item()
        if not self.delta.initialized:
            self.delta.dynamic_value = [0.0, 0.0]

    @forward
    def brightness(self, x, y, I0, delta):
        return I0 + x * delta[0] + y * delta[1]



# Traditional gradient descent with Adam
from time import time
from typing import Sequence
import torch
import numpy as np

from .base import BaseOptimizer
from .. import config
from ..models import Model

__all__ = ["Grad"]


class Grad(BaseOptimizer):
    """A gradient descent optimization wrapper for AstroPhot_Model objects.

    The default method is "NAdam", a variant of the Adam optimization algorithm.
    This optimizer uses a combination of gradient descent and Nesterov momentum for faster convergence.
    The optimizer is instantiated with a set of initial parameters and optimization options provided by the user.
    The `fit` method performs the optimization, taking a series of gradient steps until a stopping criteria is met.

    Parameters:
        model (AstroPhot_Model): an AstroPhot_Model object with which to perform optimization.
        initial_state (torch.Tensor, optional): an optional initial state for optimization.
        method (str, optional): the optimization method to use for the update step. Defaults to "NAdam".
        patience (int or None, optional): the number of iterations without improvement before the optimizer will exit early. Defaults to None.
        optim_kwargs (dict, optional): a dictionary of keyword arguments to pass to the pytorch optimizer.

    Attributes:
        model (AstroPhot_Model): the AstroPhot_Model object to optimize.
        current_state (torch.Tensor): the current state of the parameters being optimized.
        iteration (int): the number of iterations performed during the optimization.
        loss_history (list): the history of loss values at each iteration of the optimization.
        lambda_history (list): the history of parameter values at each iteration of the optimization.
        optimizer (torch.optimizer): the PyTorch optimizer object being used.
        patience (int or None): the number of iterations without improvement before the optimizer will exit early.
        method (str): the optimization method being used.
        optim_kwargs (dict): the dictionary of keyword arguments passed to the PyTorch optimizer.


    """

    def __init__(
        self,
        model: Model,
        initial_state: Sequence = None,
        likelihood="gaussian",
        patience=None,
        method="NAdam",
        optim_kwargs={},
        report_freq=10,
        **kwargs,
    ) -> None:
        """Initialize the gradient descent optimizer.

        Args:
            - model: instance of the model to be optimized.
            - initial_state: Initial state of the model.
            - patience: (optional) If a positive integer, then stop the optimization if there has been no improvement in the loss for this number of iterations.
            - method: (optional) The name of the optimization method to use. Default is NAdam.
            - optim_kwargs: (optional) Keyword arguments to be passed to the optimizer.
        """

        super().__init__(model, initial_state, **kwargs)

        self.likelihood = likelihood

        # set parameters from the user
        self.patience = patience
        self.method = method
        self.optim_kwargs = optim_kwargs
        self.report_freq = report_freq

        # Default learning rate if none given. Equal to 1 / sqrt(parames)
        if "lr" not in self.optim_kwargs:
            self.optim_kwargs["lr"] = 0.1 / (len(self.current_state) ** (0.5))

        # Instantiates the appropriate pytorch optimizer with the initial state and user provided kwargs
        self.current_state.requires_grad = True
        self.optimizer = getattr(torch.optim, self.method)(
            (self.current_state,), **self.optim_kwargs
        )

    def density(self, state: torch.Tensor) -> torch.Tensor:
        """
        Returns the density of the model at the given state vector.
        This is used to calculate the likelihood of the model at the given state.
        """
        if self.likelihood == "gaussian":
            return -self.model.gaussian_log_likelihood(state)
        elif self.likelihood == "poisson":
            return -self.model.poisson_log_likelihood(state)
        else:
            raise ValueError(f"Unknown likelihood type: {self.likelihood}")

    def step(self) -> None:
        """Take a single gradient step. Take a single gradient step.

        Computes the loss function of the model,
        computes the gradient of the parameters using automatic differentiation,
        and takes a step with the PyTorch optimizer.

        """
        self.iteration += 1

        self.optimizer.zero_grad()
        self.current_state.requires_grad = True
        loss = self.density(self.current_state)

        loss.backward()

        self.loss_history.append(loss.detach().cpu().item())
        self.lambda_history.append(np.copy(self.current_state.detach().cpu().numpy()))
        if (
            self.iteration % int(self.max_iter / self.report_freq) == 0
        ) or self.iteration == self.max_iter:
            if self.verbose > 0:
                config.logger.info(f"iter: {self.iteration}, posterior density: {loss.item():.6e}")
            if self.verbose > 1:
                config.logger.info(f"gradient: {self.current_state.grad}")
        self.optimizer.step()

    def fit(self) -> BaseOptimizer:
        """
        Perform an iterative fit of the model parameters using the specified optimizer.

        The fit procedure continues until a stopping criteria is met,
        such as the maximum number of iterations being reached,
        or no improvement being made after a specified number of iterations.

        """
        start_fit = time()
        try:
            while True:
                self.step()
                if self.iteration >= self.max_iter:
                    self.message = self.message + " fail max iteration reached"
                    break
                if (
                    self.patience is not None
                    and (len(self.loss_history) - np.argmin(self.loss_history)) > self.patience
                ):
                    self.message = self.message + " fail no improvement"
                    break
                L = np.sort(self.loss_history)
                if len(L) >= 3 and 0 < L[1] - L[0] < 1e-6 and 0 < L[2] - L[1] < 1e-6:
                    self.message = self.message + " success"
                    break
        except KeyboardInterrupt:
            self.message = self.message + " fail interrupted"

        # Set the model parameters to the best values from the fit and clear any previous model sampling
        self.model.fill_dynamic_values(
            torch.tensor(self.res(), dtype=config.DTYPE, device=config.DEVICE)
        )
        if self.verbose > 1:
            config.logger.info(
                f"Grad Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )
        return self



import unittest
import astrophot as ap
import torch
import numpy as np
from utils import make_basic_sersic, make_basic_gaussian_psf
import pytest

# torch.autograd.set_detect_anomaly(True)
######################################################################
# Model Objects
######################################################################


def test_model_sampling_modes():

    target = make_basic_sersic(90, 100)
    model = ap.Model(
        name="test sersic",
        model_type="sersic galaxy model",
        center=[40, 41.9],
        PA=60 * np.pi / 180,
        q=0.8,
        n=0.5,
        Re=20,
        Ie=1,
        target=target,
    )

    # With subpixel integration
    auto = model().data.detach().cpu().numpy()
    model.sampling_mode = "midpoint"
    midpoint = model().data.detach().cpu().numpy()
    model.sampling_mode = "simpsons"
    simpsons = model().data.detach().cpu().numpy()
    model.sampling_mode = "quad:5"
    quad5 = model().data.detach().cpu().numpy()
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    # Without subpixel integration
    model.integrate_mode = "none"
    auto = model().data.detach().cpu().numpy()
    model.sampling_mode = "midpoint"
    midpoint = model().data.detach().cpu().numpy()
    model.sampling_mode = "simpsons"
    simpsons = model().data.detach().cpu().numpy()
    model.sampling_mode = "quad:5"
    quad5 = model().data.detach().cpu().numpy()
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    # Without subpixel integration
    model.integrate_mode = "threshold"
    auto = model().data.detach().cpu().numpy()
    model.sampling_mode = "midpoint"
    midpoint = model().data.detach().cpu().numpy()
    model.sampling_mode = "simpsons"
    simpsons = model().data.detach().cpu().numpy()
    model.sampling_mode = "quad:5"
    quad5 = model().data.detach().cpu().numpy()
    assert np.allclose(midpoint, auto, rtol=1e-2), "Midpoint sampling should match auto sampling"
    assert np.allclose(midpoint, simpsons, rtol=1e-2), "Simpsons sampling should match midpoint"
    assert np.allclose(midpoint, quad5, rtol=1e-2), "Quad5 sampling should match midpoint sampling"
    assert np.allclose(simpsons, quad5, rtol=1e-6), "Quad5 sampling should match Simpsons sampling"

    model.integrate_mode = "should raise"
    with pytest.raises(ap.errors.SpecificationConflict):
        model()
    model.integrate_mode = "none"
    model.sampling_mode = "should raise"
    with pytest.raises(ap.errors.SpecificationConflict):
        model()
    model.sampling_mode = "midpoint"
    model.integrate_mode = "none"

    # test PSF modes
    model.psf = np.array([[0.05, 0.1, 0.05], [0.1, 0.4, 0.1], [0.05, 0.1, 0.05]])
    model.psf_convolve = True
    model()


def test_model_errors():

    # Target that is not a target image
    arr = torch.zeros((10, 15))
    target = ap.image.Image(data=arr, pixelscale=1.0, zeropoint=1.0)

    with pytest.raises(ap.errors.InvalidTarget):
        ap.Model(
            name="test model",
            model_type="sersic galaxy model",
            target=target,
        )

    # model that doesn't exist
    target = make_basic_sersic()
    with pytest.raises(ap.errors.UnrecognizedModel):
        ap.Model(
            name="test model",
            model_type="sersic gaaxy model",
            target=target,
        )


@pytest.mark.parametrize(
    "model_type", ap.models.ComponentModel.List_Models(usable=True, types=True)
)
def test_all_model_sample(model_type):

    target = make_basic_sersic()
    target.zeropoint = 22.5
    MODEL = ap.Model(
        name="test model",
        model_type=model_type,
        target=target,
    )
    MODEL.initialize()
    MODEL.to()
    for P in MODEL.dynamic_params:
        assert (
            P.value is not None
        ), f"Model type {model_type} parameter {P.name} should not be None after initialization"
    img = MODEL()
    assert torch.all(
        torch.isfinite(img.data)
    ), "Model should evaluate a real number for the full image"
    res = ap.fit.LM(MODEL, max_iter=10).fit()

    # sky has little freedom to fit, some more complex models need extra
    # attention to get a good fit so here we just check that they can improve
    if (
        "sky" in model_type
        or "king" in model_type
        or model_type
        in [
            "spline ray galaxy model",
            "exponential warp galaxy model",
            "spline wedge galaxy model",
            "ferrer warp galaxy model",
        ]
    ):
        assert res.loss_history[0] > res.loss_history[-1], (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )
    else:  # Most models should get significantly better after just a few iterations
        assert res.loss_history[0] > (2 * res.loss_history[-1]), (
            f"Model {model_type} should fit to the target image, but did not. "
            f"Initial loss: {res.loss_history[0]}, Final loss: {res.loss_history[-1]}"
        )

    F = MODEL.total_flux()
    assert torch.isfinite(F), "Model total flux should be finite after fitting"
    assert F > 0, "Model total flux should be positive after fitting"
    U = MODEL.total_flux_uncertainty()
    assert torch.isfinite(U), "Model total flux uncertainty should be finite after fitting"
    assert U >= 0, "Model total flux uncertainty should be non-negative after fitting"
    M = MODEL.total_magnitude()
    assert torch.isfinite(M), "Model total magnitude should be finite after fitting"
    U_M = MODEL.total_magnitude_uncertainty()
    assert torch.isfinite(U_M), "Model total magnitude uncertainty should be finite after fitting"
    assert U_M >= 0, "Model total magnitude uncertainty should be non-negative after fitting"

    allnames = set()
    for name in MODEL.build_params_array_names():
        assert name not in allnames, f"Duplicate parameter name found: {name}"
        allnames.add(name)


def test_sersic_save_load():

    target = make_basic_sersic()
    model = ap.Model(
        name="test sersic",
        model_type="sersic galaxy model",
        center=[20, 20],
        PA=60 * np.pi / 180,
        q=0.5,
        n=2,
        Re=5,
        Ie=1,
        target=target,
    )

    model.initialize()
    model.save_state("test_AstroPhot_sersic.hdf5", appendable=True)
    model.center = [30, 30]
    model.PA = 30 * np.pi / 180
    model.q = 0.8
    model.n = 3
    model.Re = 10
    model.Ie = 2
    target.crtan = [1.0, 2.0]
    model.append_state("test_AstroPhot_sersic.hdf5")
    model.load_state("test_AstroPhot_sersic.hdf5", index=0)

    assert model.center.value[0].item() == 20, "Model center should be loaded correctly"
    assert model.center.value[1].item() == 20, "Model center should be loaded correctly"
    assert model.PA.value.item() == 60 * np.pi / 180, "Model PA should be loaded correctly"
    assert model.q.value.item() == 0.5, "Model q should be loaded correctly"
    assert model.n.value.item() == 2, "Model n should be loaded correctly"
    assert model.Re.value.item() == 5, "Model Re should be loaded correctly"
    assert model.Ie.value.item() == 1, "Model Ie should be loaded correctly"
    assert model.target.crtan.value[0] == 0.0, "Model target crtan should be loaded correctly"
    assert model.target.crtan.value[1] == 0.0, "Model target crtan should be loaded correctly"


@pytest.mark.parametrize("center", [[20, 20], [25.1, 17.324567]])
@pytest.mark.parametrize("PA", [0, 60 * np.pi / 180])
@pytest.mark.parametrize("q", [0.2, 0.8])
@pytest.mark.parametrize("n", [1, 4])
@pytest.mark.parametrize("Re", [10, 25.1])
def test_chunk_sample(center, PA, q, n, Re):
    target = make_basic_sersic()
    model = ap.Model(
        name="test sersic",
        model_type="sersic galaxy model",
        center=center,
        PA=PA,
        q=q,
        n=n,
        Re=Re,
        Ie=10.0,
        target=target,
        integrate_mode="none",
    )

    full_img = model.sample()

    chunk_img = target.model_image()

    for chunk in model.window.chunk(20**2):
        sample = model.sample(window=chunk)
        chunk_img += sample

    assert torch.allclose(
        full_img.data, chunk_img.data
    ), "Chunked sample should match full sample within tolerance"
